from pathlib import Path
import sys

import alembic.config

here = Path(__file__).parent


def main() -> None:
    argv = [
        # Use our custom config path
        "--config",
        str(here / "alembic.ini"),
        "--raiseerr",
        "upgrade",
        "head",
    ]
    alembic.config.main(argv=argv)  # type: ignore



"""Use cases for devices."""

from fastapi import HTTPException

from energy_assistant.EmhassOptimizer import EmhassOptimizer
from energy_assistant.models.forecast import ForecastSchema

from .schema import CreateModelResponse, TuneModelResponse


class ReadForecast:
    """Read the forecast."""

    async def execute(self, optimizer: EmhassOptimizer) -> ForecastSchema:
        """Execute the read all devices use case."""
        if optimizer is not None:
            return optimizer.get_forecast()


class CreateModel:
    """Create the forecast model."""

    async def execute(
        self, days_to_retrieve: int, optimizer: EmhassOptimizer
    ) -> CreateModelResponse:
        """Execute the create model use case."""
        try:
            r2 = optimizer.forecast_model_fit(False, days_to_retrieve)
            return CreateModelResponse(r2=r2)
        except UnboundLocalError:
            raise HTTPException(status_code=400, detail="Creation of the model failed.")


class TuneModel:
    """Create the forecast model."""

    async def execute(self, optimizer: EmhassOptimizer) -> TuneModelResponse:
        """Execute the create model use case."""
        optimizer.forecast_model_tune()
        return TuneModelResponse(model="Tuned")



"""Tests for the devices."""

import uuid
from datetime import date

import pytest
from sqlalchemy.ext.asyncio import AsyncSession

from energy_assistant.devices import EnergyIntegrator, PowerModes
from energy_assistant.devices.home import Home
from energy_assistant.devices.homeassistant import HomeassistantDevice
from energy_assistant.devices.registry import DeviceTypeRegistry
from energy_assistant.models.home import HomeMeasurement
from energy_assistant.storage import Database, session_storage


async def setup_data(session: AsyncSession) -> None:
    """Set up the data in the database."""
    from energy_assistant.models.device import Device, DeviceMeasurement
    from energy_assistant.models.home import HomeMeasurement

    device1 = Device(
        id=uuid.UUID("1a8ac2d6-5695-427a-a3c5-ef567b34e5ec"),
        name="Device 1",
        icon="mdi-home",
        power_mode=PowerModes.DEVICE_CONTROLLED,
    )
    session.add(device1)
    device2 = Device(
        id=uuid.UUID("2a8ac2d6-5695-427a-a3c5-ef567b34e5ec"),
        name="Device 2",
        icon="mdi-home",
        power_mode=PowerModes.DEVICE_CONTROLLED,
    )
    session.add(device2)
    await session.flush()
    devices = [device1, device2]

    dates = [date(2023, 1, 9), date(2023, 1, 10), date(2023, 1, 11)]
    for m, d in enumerate(dates):
        home_measurement = HomeMeasurement(
            name="my home",
            measurement_date=d,
            consumed_energy=123 + m,
            solar_consumed_energy=120 + 0.5 * m,
            solar_produced_energy=150 + m,
            grid_imported_energy=1123 + m,
            grid_exported_energy=1540 + m,
            device_measurements=[],
        )
        session.add(home_measurement)
        await session.flush()
        for i, device in enumerate(devices):
            device_measurement = DeviceMeasurement(
                home_measurement_id=home_measurement.id,
                consumed_energy=2 + 0.5 * i + m,
                solar_consumed_energy=1 + i + 0.5 * m,
                device_id=device.id,
            )
            session.add(device_measurement)
        await session.flush()

    await session.commit()


def test_integrator() -> None:
    """Test the energy integrator."""
    integrator = EnergyIntegrator()
    integrator.restore_state(15, 20)
    integrator.add_measurement(30, 0.1)
    assert integrator.consumed_solar_energy == 16


@pytest.mark.asyncio
async def test_load(session: AsyncSession, device_type_registry: DeviceTypeRegistry) -> None:
    """Test the loading of the devices."""
    await setup_data(session)

    home = Home(
        {
            "name": "my home",
            "solar_power": "solar_power_id",
            "grid_supply_power": "grid_supply_power_id",
            "solar_energy": "solar_energy_id",
            "imported_energy": "imported_energy_id",
            "exported_energy": "exported_energy_id",
        },
        session_storage,
        device_type_registry,
    )
    device = HomeassistantDevice(
        {
            "id": "1a8ac2d6-5695-427a-a3c5-ef567b34e5ec",
            "name": "Device 1",
            "power": "power_id",
            "energy": "energy_id",
        },
        session_storage,
        device_type_registry,
    )
    home.add_device(device)
    db = Database()
    home_measurement = await HomeMeasurement.read_last(session, True)
    assert home_measurement is not None
    await db.restore_home_state(home, session)
    assert home.consumed_energy == 125
    assert home.consumed_solar_energy == 121
    assert home.devices[0].consumed_solar_energy == 2
    assert home.devices[0].consumed_energy == 4

    assert True



"""Schemas for home measurement api."""

from pydantic import BaseModel


class CreateModelResponse(BaseModel):
    """API Request for create the forecast model."""

    r2: float


class TuneModelResponse(BaseModel):
    """API Request for tuning the forecast model."""

    model: str



"""Device type registry for device data."""

import logging
import os
from dataclasses import dataclass

import yaml

from energy_assistant.constants import (
    DEFAULT_NOMINAL_DURATION,
    DEFAULT_NOMINAL_POWER,
    ROOT_LOGGER_NAME,
)

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)


@dataclass(frozen=True, eq=True)
class DeviceTypeId:
    """The of a device type."""

    manufacturer: str
    model: str


@dataclass
class DeviceType:
    """The device type of a device."""

    icon: str
    nominal_power: float
    nominal_duration: float
    constant: bool
    state_on_threshold: float
    state_off_threshold: float
    state_off_upper: float
    state_off_lower: float
    state_off_for: float
    trailing_zeros_for: float


class DeviceTypeRegistry:
    """The registry for device types."""

    def __init__(self) -> None:
        """Create a Device Type Registry instance."""
        self._registry: dict[DeviceTypeId, DeviceType] = {}

    def load(self, config_folder: str) -> None:
        """Load the registry from the configuration folder."""
        for root, subdirs, files in os.walk(config_folder):
            for file in files:
                file_path = os.path.join(root, file)
                self.load_device_type_file(file_path)
            for subdir in subdirs:
                self.load(os.path.join(root, subdir))

    def load_device_type_file(self, filename: str) -> None:
        """Load a device type config file and add it to the registry."""
        with open(filename) as stream:
            try:
                config = yaml.safe_load(stream)
            except yaml.YAMLError as ex:
                LOGGER.error(ex)
            except Exception as ex:
                LOGGER.error(ex)
            else:
                device_type_config = config.get("device_type")
                if device_type_config is None:
                    LOGGER.error(
                        f"Device type config file {filename} does not contain a device_type item."
                    )
                else:
                    manufacturer = device_type_config.get("manufacturer")
                    model = device_type_config.get("model")
                    icon = device_type_config.get("icon")
                    nominal_power = device_type_config.get("nominal_power", DEFAULT_NOMINAL_POWER)
                    nominal_duration = device_type_config.get(
                        "nominal_duration", DEFAULT_NOMINAL_DURATION
                    )
                    constant = device_type_config.get("constant", False)
                    if model is None or manufacturer is None or icon is None:
                        LOGGER.error(
                            f"Manufacturer or Model or Icon not set in device type config file {filename}"
                        )
                    else:
                        device_state_config = device_type_config.get("state")
                        state_on_threshold: float | None = None
                        state_on_config = device_state_config.get("state_on")
                        if state_on_config is not None:
                            state_on_threshold = state_on_config.get("threshold")

                        state_off_threshold: float = 0.0
                        state_off_upper: float | None = None
                        state_off_lower: float | None = None
                        state_off_for: float | None = None
                        state_off_config = device_state_config.get("state_off")
                        if state_off_config is not None:
                            state_off_threshold = state_off_config.get("threshold", 0.0)
                            state_off_upper = state_off_config.get("upper")
                            state_off_lower = state_off_config.get("lower")
                            state_off_for = state_off_config.get("for")
                            trailing_zeros_for = state_off_config.get("trailing_zeros_for")
                        if (
                            state_on_threshold is not None
                            and state_off_upper is not None
                            and state_off_lower is not None
                            and state_off_for is not None
                            and trailing_zeros_for is not None
                        ):
                            device_type = DeviceType(
                                icon=icon,
                                nominal_power=nominal_power,
                                nominal_duration=nominal_duration,
                                constant=constant,
                                state_on_threshold=state_on_threshold,
                                state_off_threshold=state_off_threshold,
                                state_off_upper=state_off_upper,
                                state_off_lower=state_off_lower,
                                state_off_for=state_off_for,
                                trailing_zeros_for=trailing_zeros_for,
                            )
                            self._registry[DeviceTypeId(manufacturer=manufacturer, model=model)] = (
                                device_type
                            )

    def get_device_type(self, manufacturer: str, model: str) -> DeviceType | None:
        """Get the device type for a given manufacturer and model."""
        return self._registry.get(DeviceTypeId(manufacturer=manufacturer, model=model))



"""EVCC Devices."""

from energy_assistant import Optimizer
from energy_assistant.devices.analysis import DataBuffer
from energy_assistant.mqtt import MQTT_CHANNEL

from . import (
    LoadInfo,
    OnOffState,
    PowerModes,
    SessionStorage,
    State,
    StateId,
    StatesRepository,
)
from .config import get_config_param
from .device import DeviceWithState


class EvccDevice(DeviceWithState):
    """Evcc load points as devices."""

    def __init__(self, config: dict, session_storage: SessionStorage):
        """Create a Stiebel Eltron heatpump."""
        super().__init__(config, session_storage)
        self._evcc_topic: str = get_config_param(config, "evcc_topic")
        self._loadpoint_id: int = int(get_config_param(config, "load_point_id"))
        self._is_continous: bool = bool(config.get("continuous", True))
        self._nominal_power: float | None = config.get("nominal_power")
        self._state = OnOffState.UNKNOWN
        self._power: State | None = None
        self._consumed_energy: State | None = None
        self._mode: State | None = None
        self._vehicle_soc: State | None = None
        self._max_current: State | None = None
        self._is_connected: State | None = None
        self._supported_power_modes = [
            PowerModes.DEVICE_CONTROLLED,
            PowerModes.OFF,
            PowerModes.PV,
            PowerModes.MIN_PV,
            PowerModes.FAST,
            PowerModes.OPTIMIZED,
        ]
        self._utility_meter = self.add_utility_meter("energy")

    def get_device_topic_id(self, name: str) -> StateId:
        """Get the id of a topic of this load point."""
        return StateId(
            id=f"{self._evcc_topic}/loadpoints/{self._loadpoint_id}/{name}",
            channel=MQTT_CHANNEL,
        )

    @property
    def type(self) -> str:
        """The device type."""
        return "evcc"

    async def update_state(
        self, state_repository: StatesRepository, self_sufficiency: float
    ) -> None:
        """Update the state of the Stiebel Eltron device."""
        old_state = self.state == OnOffState.ON
        charging = state_repository.get_state(self.get_device_topic_id("charging"))
        if charging is not None:
            self._state = OnOffState.ON if charging.value == "true" else OnOffState.OFF
        else:
            self._state = OnOffState.UNKNOWN
        new_state = self.state == OnOffState.ON

        self._consumed_energy = state_repository.get_state(
            self.get_device_topic_id("chargeTotalImport")
        )
        if self.consumed_energy == 0:
            state = state_repository.get_state(self.get_device_topic_id("sessionEnergy"))
            if state is not None:
                self._consumed_energy = self._utility_meter.update_energy_state(state)
        self._consumed_solar_energy.add_measurement(self.consumed_energy, self_sufficiency)
        self._power = state_repository.get_state(self.get_device_topic_id("chargePower"))
        self._mode = state_repository.get_state(self.get_device_topic_id("mode"))
        self._vehicle_soc = state_repository.get_state(self.get_device_topic_id("vehicleSoc"))
        self._vehicle_capacity = state_repository.get_state(
            self.get_device_topic_id("vehicleCapacity")
        )
        self._max_current = state_repository.get_state(self.get_device_topic_id("maxCurrent"))
        self._is_connected = state_repository.get_state(self.get_device_topic_id("connected"))

        if self._energy_snapshot is None:
            self.set_snapshot(self.consumed_solar_energy, self.consumed_energy)

        await self.update_session(old_state, new_state, "EVCC")

    async def update_power_consumption(
        self,
        state_repository: StatesRepository,
        optimizer: Optimizer,
        grid_exported_power_data: DataBuffer,
    ) -> None:
        """Update the device based on the current pv availability."""
        new_state = ""
        if self.power_mode == PowerModes.OFF:
            new_state = "off"
        elif self.power_mode == PowerModes.PV:
            new_state = "pv"
        elif self.power_mode == PowerModes.MIN_PV:
            new_state = "minpv"
        elif self.power_mode == PowerModes.FAST:
            new_state = "now"
        elif self.power_mode == PowerModes.OPTIMIZED:
            # TODO: Implement Optimized with emhass
            new_state = "pv"
        if new_state != self._mode:
            state_repository.set_state(self.get_device_topic_id("mode/set"), new_state)

    @property
    def evcc_mqtt_subscription_topic(self) -> str:
        """Get the evvc base topic for the subscription."""
        return f"{self._evcc_topic}/#"

    @property
    def consumed_energy(self) -> float:
        """Consumed energy in kWh."""
        return self._consumed_energy.numeric_value if self._consumed_energy else 0.0

    @property
    def state(self) -> str:
        """The state of the device. The state is `on` in case the device is loading."""
        return self._state

    @property
    def power(self) -> float:
        """Current power consumption of the device."""
        return self._power.numeric_value if self._power else 0.0

    @property
    def mode(self) -> str:
        """Current PV mode of the device."""
        return self._mode.value if self._mode else "unknown"

    @property
    def vehicle_soc(self) -> float:
        """State of Charge of the connected vehicle."""
        return self._vehicle_soc.numeric_value if self._vehicle_soc else 0.0

    @property
    def vehicle_capacity(self) -> float:
        """Capacity of the connected vehicle."""
        return self._vehicle_capacity.numeric_value if self._vehicle_capacity else 0.0

    @property
    def icon(self) -> str:
        """Icon of the device."""
        return "mdi-car-electric"

    @property
    def available(self) -> bool:
        """Check if the device is available."""
        return True

    def restore_state(self, consumed_solar_energy: float, consumed_energy: float) -> None:
        """Restore the previously stored state."""
        super().restore_state(consumed_solar_energy, consumed_energy)
        self._consumed_energy = State(
            self.get_device_topic_id("chargeTotalImport").id, str(consumed_energy)
        )

    @property
    def attributes(self) -> dict[str, str]:
        """Get the attributes of the device for the UI."""
        result: dict[str, str] = {
            **super().attributes,
            "pv_mode": self.mode,
        }
        if self._vehicle_soc is not None:
            result["vehicle_soc"] = f"{round(self.vehicle_soc)} %"
        return result

    def get_load_info(self) -> LoadInfo | None:
        """Get the current deferrable load info."""
        if (
            self._is_connected is not None
            and self._max_current is not None
            and self._is_connected.value == "true"
        ):
            if self.state == OnOffState.ON:
                remainingEnergy = (1 - self.vehicle_soc / 100) * self.vehicle_capacity * 1000
                if remainingEnergy > 0:
                    power: float = (
                        self._nominal_power
                        if self._nominal_power is not None
                        else (self._max_current.numeric_value * 230)
                    )  # TODO: Multiply with active phases
                    return LoadInfo(
                        device_id=self.id,
                        nominal_power=power,
                        duration=remainingEnergy / power * 3600,
                        is_continous=self._is_continous,
                        is_deferrable=self.power_mode == PowerModes.OPTIMIZED,
                    )
        return None



"""Use cases for devices."""

import logging
import uuid
from typing import AsyncIterator

from fastapi import HTTPException

from energy_assistant.constants import ROOT_LOGGER_NAME
from energy_assistant.db import AsyncSession
from energy_assistant.devices import PowerModes
from energy_assistant.devices.home import Home
from energy_assistant.models.device import Device, DeviceMeasurement
from energy_assistant.models.schema import DeviceMeasurementSchema, DeviceSchema

from . import OTHER_DEVICE

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)


class ReadAllDevices:
    """Read all devices use case."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a read all devices use case."""
        self.async_session = session

    async def execute(
        self, home: Home | None, filter_with_session_log_enties: bool
    ) -> AsyncIterator[DeviceSchema]:
        """Execute the read all devices use case."""
        async with self.async_session() as session:
            async for device in Device.read_all(session, False, filter_with_session_log_enties):
                if not filter_with_session_log_enties or len(device.session_log_entries) > 0:
                    if device.type is not None:  # TODO: Remove this at a late point
                        result = DeviceSchema.model_validate(device)
                        d = home.get_device(device.id) if home is not None else None
                        if d is not None:
                            result.supported_power_modes = list(d.supported_power_modes)
                            result.power_mode = d.power_mode
                        yield result
        if not filter_with_session_log_enties:
            yield DeviceSchema(
                id=OTHER_DEVICE,
                name="Other",
                icon="mdi-home",
                type="other",
                config="",
                supported_power_modes=[],
                power_mode="",
            )


class ReadDevice:
    """Read a device use case."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a read device use case."""
        self.async_session = session

    async def execute(self, device_id: uuid.UUID, home: Home) -> DeviceSchema:
        """Execute the read device use case."""
        async with self.async_session() as session:
            device = await Device.read_by_id(session, device_id)
            if not device:
                raise HTTPException(status_code=404)
            result = DeviceSchema.model_validate(device)
            d = home.get_device(device.id)
            if d is not None:
                result.supported_power_modes = list(d.supported_power_modes)
                result.power_mode = d.power_mode
            return result


class ReadDeviceMeasurements:
    """Read the device measurements use case."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a read device use case."""
        self.async_session = session

    async def execute(self, device_id: uuid.UUID) -> AsyncIterator[DeviceMeasurementSchema]:
        """Execute the read device use case."""
        async with self.async_session() as session:
            async for device_measurement in DeviceMeasurement.read_by_device_id(session, device_id):
                yield DeviceMeasurementSchema.model_validate(device_measurement)


class UpdateDevicePowerMode:
    """Update the power mode of  a device use case."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a uodate device power mode use case."""
        self.async_session = session

    async def execute(self, device_id: uuid.UUID, power_mode: str, home: Home) -> DeviceSchema:
        """Execute the update device power nmode use case."""
        async with self.async_session.begin() as session:
            d = home.get_device(device_id)
            if d is None:
                raise HTTPException(status_code=404)
            try:
                d.set_power_mode(PowerModes[power_mode.upper()])
                device = await Device.read_by_id(session, device_id)
                if not device:
                    raise HTTPException(status_code=404)

                if device.type is None:
                    raise HTTPException(status_code=500)

                await device.update(
                    session, device.name, device.icon, power_mode, device.type, device.get_config()
                )
                await session.refresh(device)
                result = DeviceSchema.model_validate(device)
                result.supported_power_modes = list(d.supported_power_modes)
                result.power_mode = d.power_mode
                return result
            except Exception:
                LOGGER.error("Invalid power mode: " + power_mode)
                raise HTTPException(status_code=404)


class DeleteDevice:
    """Delete a device use case."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a delete a device use case."""
        self.async_session = session

    async def execute(self, device_id: uuid.UUID, home: Home | None) -> None:
        """Execute the delete a device use case."""
        async with self.async_session.begin() as session:
            device = await Device.read_by_id(session, device_id)
            if not device:
                return
            await Device.delete(session, device)
            if home is not None:
                home.remove_device(device_id)



"""Stiebel Eltron device implementation."""

from energy_assistant import Optimizer
from energy_assistant.constants import POWER_HYSTERESIS
from energy_assistant.devices.device import DeviceWithState
from energy_assistant.devices.state_value import StateValue

from . import LoadInfo, PowerModes, SessionStorage, State, StateId, StatesRepository
from .analysis import DataBuffer
from .config import DeviceConfigException, get_config_param
from .homeassistant import HOMEASSISTANT_CHANNEL, assign_if_available

DEFAULT_NOMINAL_POWER = 5000


def numeric_value(value: str | None) -> float | None:
    """Convert into a number."""
    if value is None:
        return None
    try:
        return float(value)
    except ValueError:
        return None


class HeatPumpDevice(DeviceWithState):
    """Stiebel Eltron heatpump. This can be either a water heating part or a heating part."""

    def __init__(self, config: dict, session_storage: SessionStorage):
        """Create a Stiebel Eltron heatpump."""
        super().__init__(config, session_storage)
        energy_config = config.get("energy")
        if energy_config is not None:
            self._consumed_energy_value = StateValue(energy_config)
        else:
            raise DeviceConfigException("Parameter energy is missing in the configuration")
        self._actual_temp_entity_id: str = get_config_param(config, "temperature")
        self._actual_temp: State | None = None
        self._state: State | None = None
        self._nominal_power: float = config.get("nominal_power", DEFAULT_NOMINAL_POWER)
        self._comfort_target_temperature_entity_id = config.get("comfort_target_temperature")
        self._target_temperature_normal: float | None = numeric_value(
            config.get("target_temperature_normal")
        )
        self._target_temperature_pv: float | None = numeric_value(
            config.get("target_temperatrure_pv")
        )
        if (
            self._target_temperature_normal is not None
            and self._target_temperature_pv is not None
            and self._comfort_target_temperature_entity_id is not None
        ):
            self._supported_power_modes.append(PowerModes.PV)
            self.supported_power_modes.append(PowerModes.OPTIMIZED)

        self._state_entity_id: str = get_config_param(config, "state")
        self._consumed_energy: State | None = None
        self._icon = "mdi-heat-pump"

    @property
    def type(self) -> str:
        """The device type."""
        return "heat-pump"

    async def update_state(
        self, state_repository: StatesRepository, self_sufficiency: float
    ) -> None:
        """Update the state of the Stiebel Eltron device."""
        old_state = self.state == "on"
        self._state = assign_if_available(
            self._state, state_repository.get_state(self._state_entity_id)
        )
        new_state = self.state == "on"

        self._consumed_energy = assign_if_available(
            self._consumed_energy,
            self._consumed_energy_value.evaluate(state_repository),
        )
        self._consumed_solar_energy.add_measurement(self.consumed_energy, self_sufficiency)
        self._actual_temp = assign_if_available(
            self._actual_temp, state_repository.get_state(self._actual_temp_entity_id)
        )

        if self._energy_snapshot is None:
            self.set_snapshot(self.consumed_solar_energy, self.consumed_energy)

        await self.update_session(old_state, new_state, "Water heater")

    async def update_power_consumption(
        self,
        state_repository: StatesRepository,
        optimizer: Optimizer,
        grid_exported_power_data: DataBuffer,
    ) -> None:
        """Update the device based on the current pv availability."""
        if (
            self._target_temperature_normal is not None
            and self._target_temperature_pv is not None
            and self._comfort_target_temperature_entity_id is not None
        ):
            current_target_temperature = state_repository.get_state(
                self._comfort_target_temperature_entity_id
            )
            if current_target_temperature is not None:
                target_temperature: float = current_target_temperature.numeric_value
                if self.power_mode == PowerModes.PV:
                    if self.state == "off":
                        avg_300 = grid_exported_power_data.get_average_for(300)
                        if avg_300 > self.requested_additional_power * (1 + POWER_HYSTERESIS):
                            target_temperature = self._target_temperature_pv
                        elif avg_300 < self.requested_additional_power * (1 - POWER_HYSTERESIS):
                            target_temperature = self._target_temperature_normal
                elif self.power_mode == PowerModes.OPTIMIZED:
                    target_temperature = self._get_temperature_for_optimized(
                        optimizer, target_temperature
                    )
                if target_temperature != current_target_temperature.numeric_value:
                    state_repository.set_state(
                        StateId(
                            id=self._comfort_target_temperature_entity_id,
                            channel=HOMEASSISTANT_CHANNEL,
                        ),
                        str(target_temperature),
                    )

    def _get_temperature_for_optimized(
        self, optimizer: Optimizer, current_target_temperature: float
    ) -> float:
        if (
            self.state == "off"
            and self._target_temperature_normal is not None
            and self._target_temperature_pv is not None
        ):
            power = optimizer.get_optimized_power(self._id)
            if power > 0:
                return self._target_temperature_pv
            else:
                return self._target_temperature_normal
        else:
            return current_target_temperature

    def get_load_info(self) -> LoadInfo | None:
        """Get the current deferrable load info."""
        if self.power_mode == PowerModes.OPTIMIZED:
            return LoadInfo(
                device_id=self.id,
                nominal_power=self._nominal_power,
                duration=1800,  # 0.5h -> TODO: make this configurable
                is_continous=False,
                is_deferrable=True,
            )
        return None

    @property
    def requested_additional_power(self) -> float:
        """How much power the device could consume in pv mode."""
        # TODO: Consider the actual temperature
        return self._nominal_power if self.state == "off" else 0.0

    @property
    def consumed_energy(self) -> float:
        """Consumed energy in kWh."""
        return self._consumed_energy.numeric_value if self._consumed_energy else 0.0

    @property
    def state(self) -> str:
        """The state of the device. The state is `on` in case the device is heating."""
        return self._state.value if self._state and self._state.value else "unknown"

    @property
    def power(self) -> float:
        """Current power consumption of the device."""
        if self._state is not None:
            return self._nominal_power if self._state.value == "on" else 0.0
        else:
            return 0.0

    @property
    def actual_temperature(self) -> float:
        """The actual temperature of the heating or water."""
        return self._actual_temp.numeric_value if self._actual_temp else 0.0

    @property
    def icon(self) -> str:
        """Icon of the device."""
        return self._icon

    @property
    def available(self) -> bool:
        """Is the device available?."""
        return (
            self._consumed_energy is not None
            and self._consumed_energy.available
            and self._actual_temp is not None
            and self._actual_temp.available
            and self._state is not None
            and self._state.available
        )

    def restore_state(self, consumed_solar_energy: float, consumed_energy: float) -> None:
        """Restore the previously stored state."""
        super().restore_state(consumed_solar_energy, consumed_energy)
        self._consumed_energy = State("", str(consumed_energy))

    @property
    def attributes(self) -> dict[str, str]:
        """Get the attributes of the device for the UI."""
        result: dict[str, str] = {
            **super().attributes,
            "actual_temperature": f"{self.actual_temperature} °C",
        }
        return result



"""Interface to the homeassistant instance."""

import logging

import requests  # type: ignore

from energy_assistant import Optimizer
from energy_assistant.constants import (
    DEFAULT_NOMINAL_DURATION,
    DEFAULT_NOMINAL_POWER,
    POWER_HYSTERESIS,
    ROOT_LOGGER_NAME,
)
from energy_assistant.devices.analysis import DataBuffer
from energy_assistant.devices.registry import DeviceType, DeviceTypeRegistry
from energy_assistant.devices.state_value import StateValue

from . import (
    LoadInfo,
    Location,
    OnOffState,
    PowerModes,
    SessionStorage,
    State,
    StateId,
    StatesRepository,
    StatesSingleRepository,
    assign_if_available,
)
from .config import DeviceConfigException, get_config_param
from .device import DeviceWithState

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)
UNAVAILABLE = "unavailable"

HOMEASSISTANT_CHANNEL = "ha"


class HomeassistantState(State):
    """Abstract base class for states."""

    def __init__(self, id: str, value: str, attributes: dict = {}) -> None:
        """Create a State instance."""
        super().__init__(id, value, attributes)

        if value == UNAVAILABLE:
            self._available = False
        else:
            self._available = True

    @property
    def name(self) -> str:
        """The name of the State."""
        if self._attributes is not None:
            return str(self._attributes.get("friendly_name"))
        return self._id

    @property
    def unit(self) -> str:
        """Unit of the state."""
        if self._attributes is not None:
            return str(self._attributes.get("unit_of_measurement"))
        return ""


class Homeassistant(StatesSingleRepository):
    """Home assistant proxy."""

    def __init__(self, url: str, token: str, demo_mode: bool) -> None:
        """Create an instance of the Homeassistant class."""
        super().__init__(HOMEASSISTANT_CHANNEL)
        self._url = url
        self._token = token
        self._demo_mode = demo_mode is not None and demo_mode

    @property
    def url(self) -> str:
        """URL of the home assistant instance."""
        return self._url

    @property
    def token(self) -> str:
        """Token of the home assistant instance."""
        return self._token

    def read_states(self) -> None:
        """Read the states from the homeassistant instance."""
        if self._demo_mode:
            self._read_states["sensor.solaredge_i1_ac_power"] = HomeassistantState(
                "sensor.solaredge_i1_ac_power", "10000"
            )
            self._read_states["sensor.solaredge_m1_ac_power"] = HomeassistantState(
                "sensor.solaredge_m1_ac_power", "6000"
            )
            self._read_states["sensor.keba_charge_power"] = HomeassistantState(
                "sensor.keba_charge_power", "2500"
            )
            self._read_states["sensor.tumbler_power"] = HomeassistantState(
                "sensor.tumbler_power", "600"
            )
            self._read_states["sensor.officedesk_power"] = HomeassistantState(
                "sensor.officedesk_power", "40"
            )
            self._read_states["sensor.rack_power"] = HomeassistantState("sensor.rack_power", "80")
        else:
            headers = {
                "Authorization": f"Bearer {self._token}",
                "content-type": "application/json",
            }
            try:
                response = requests.get(f"{self._url}/api/states", headers=headers)

                if response.ok:
                    states = response.json()
                    self._read_states.clear()
                    for state in states:
                        entity_id = state.get("entity_id")
                        self._read_states[entity_id] = HomeassistantState(
                            entity_id, state.get("state"), state.get("attributes")
                        )
                    self._template_states = None

            except Exception:
                LOGGER.exception("Exception during homeassistant update_states: ")

    def write_states(self) -> None:
        """Send the changed states to hass."""
        if not self._demo_mode:
            headers = {
                "Authorization": f"Bearer {self._token}",
                "content-type": "application/json",
            }
            try:
                for id, state in self._write_states.items():
                    if id.startswith("number"):
                        data = {"entity_id": id, "value": state.value}
                        response = requests.post(
                            f"{self._url}/api/services/number/set_value",
                            headers=headers,
                            json=data,
                        )
                        if not response.ok:
                            LOGGER.error("State update in hass failed")
                    elif id.startswith("switch"):
                        data = {"entity_id": id}
                        response = requests.post(
                            f"{self._url}/api/services/switch/turn_{state.value}",
                            headers=headers,
                            json=data,
                        )
                        if not response.ok:
                            LOGGER.error("Turn switch update in hass failed")
                    elif id.startswith("sensor"):
                        sensor_data: dict = {
                            "state": state.value,
                            "attributes": state.attributes,
                        }
                        response = requests.post(
                            f"{self._url}/api/states/{id}",
                            headers=headers,
                            json=sensor_data,
                        )
                        if not response.ok:
                            LOGGER.error(f"State update for {id} in hass failed")
                    else:
                        LOGGER.error(f"Writing to id {id} is not yet implemented.")
            except Exception:
                LOGGER.exception("Exception during homeassistant update_states.")
            self._write_states.clear()

    def get_config(self) -> dict:
        """Read the Homeassistant configuration."""
        headers = {
            "Authorization": f"Bearer {self._token}",
            "content-type": "application/json",
        }
        response = requests.get(f"{self._url}/api/config", headers=headers)

        if response.ok:
            return response.json()
        else:
            raise Exception("Could not get location from Home Assistant.")

    def get_location(self) -> Location:
        """Read the location from the Homeassistant configuration."""
        config = self.get_config()

        return Location(
            latitude=config.get("latitude", ""),
            longitude=config.get("longitude", ""),
            elevation=config.get("elevation", ""),
            time_zone=config.get("time_zone", ""),
        )


class HomeassistantDevice(DeviceWithState):
    """A generic Homeassistant device."""

    def __init__(
        self,
        config: dict,
        session_storage: SessionStorage,
        device_type_registry: DeviceTypeRegistry,
    ) -> None:
        """Create a generic Homeassistant device."""
        super().__init__(config, session_storage)
        self._power_entity_id: str = get_config_param(config, "power")
        energy_config = config.get("energy")
        if energy_config is not None:
            self._consumed_energy_value = StateValue(energy_config)
        else:
            raise DeviceConfigException("Parameter energy is missing in the configuration")

        energy_scale: float | None = config.get("energy_scale")
        if energy_scale is not None:
            self._consumed_energy_value.set_scale(energy_scale)
            LOGGER.warning(
                f"Homeassistant device with id {self.id} is configured with energy_scale. This is deprecated and will no longer be supported."
            )

        self._output_id: str | None = config.get("output")
        self._nominal_power: float | None = None
        self._nominal_duration: float | None = None
        self._is_constant: bool | None = None

        self._power: State | None = None
        self._consumed_energy: State | None = None
        self._output_state: State | None = None
        self._icon: str = str(config.get("icon", "mdi-home"))

        if self._output_id is not None:
            self._supported_power_modes.append(PowerModes.PV)
            self.supported_power_modes.append(PowerModes.OPTIMIZED)

        self._power_data = DataBuffer()
        manufacturer = config.get("manufacturer")
        model = config.get("model")
        self._device_type: DeviceType | None = None
        if model is not None and manufacturer is not None:
            self._device_type = device_type_registry.get_device_type(manufacturer, model)
            self._nominal_power = config.get(
                "nominal_power", self._device_type.nominal_power if self._device_type else None
            )
            self._nominal_duration = config.get(
                "nominal_duration",
                self._device_type.nominal_duration if self._device_type else None,
            )
            self._is_constant = config.get(
                "constant", self._device_type.constant if self._device_type else None
            )
        else:
            self._nominal_power = config.get("nominal_power")
            self._nominal_duration = config.get("nominal_duration")
            self._is_constant = config.get("constant")

        state_detection: dict = config.get("state", {})
        if self._device_type is None and state_detection:
            state_on = state_detection.get("state_on", {})
            state_off = state_detection.get("state_off", {})
            self._device_type = DeviceType(
                str(config.get("icon", "mdi:lightning-bolt")),
                self._nominal_power if self._nominal_power is not None else DEFAULT_NOMINAL_POWER,
                (
                    self._nominal_duration
                    if self._nominal_duration is not None
                    else DEFAULT_NOMINAL_DURATION
                ),
                self._is_constant if self._is_constant is not None else False,
                state_on.get("threshold", 2),
                state_off.get("threshold", 0),
                state_off.get("upper", 0),
                state_off.get("lower", 0),
                state_off.get("for", 0),
                state_off.get("trailing_zeros_for", 10),
            )
        self._state: str
        if self.has_state:
            self._state = OnOffState.UNKNOWN
        else:
            self._state = "not supported"

    @property
    def type(self) -> str:
        """The device type."""
        return "homeassistant"

    async def update_state(
        self, state_repository: StatesRepository, self_sufficiency: float
    ) -> None:
        """Update the own state from the states of a StatesRepository."""
        if self._output_id is not None:
            self._output_state = assign_if_available(
                self._output_state, state_repository.get_state(self._output_id)
            )
        else:
            self._output_state = None
        self._power = assign_if_available(
            self._power, state_repository.get_state(self._power_entity_id)
        )
        self._consumed_energy = assign_if_available(
            self._consumed_energy,
            self._consumed_energy_value.evaluate(state_repository),
        )
        self._consumed_solar_energy.add_measurement(self.consumed_energy, self_sufficiency)
        if self._energy_snapshot is None:
            self.set_snapshot(self.consumed_solar_energy, self.consumed_energy)

        if self.has_state:
            old_state = self.state == OnOffState.ON
            if self._device_type is not None:
                self._power_data.add_data_point(self.power)
                if (
                    self.state != OnOffState.ON
                    and self.power > self._device_type.state_on_threshold
                ):
                    self._state = OnOffState.ON
                elif self.state != OnOffState.OFF:
                    if (
                        self.state == OnOffState.ON
                        and self.power <= self._device_type.state_off_threshold
                    ):
                        is_between = (
                            self._device_type.state_off_for > 0
                            and self._power_data.is_between(
                                self._device_type.state_off_lower,
                                self._device_type.state_off_upper,
                                self._device_type.state_off_for,
                                without_trailing_zeros=True,
                            )
                        )
                        max = (
                            self._device_type.trailing_zeros_for > 0
                            and self._power_data.get_max_for(self._device_type.trailing_zeros_for)
                        )
                        if is_between or max <= self._device_type.state_off_threshold:
                            self._state = OnOffState.OFF
                    elif self.state == OnOffState.UNKNOWN:
                        self._state = OnOffState.OFF
            new_state = self.state == OnOffState.ON
            await super().update_session(old_state, new_state, "Power State Device")

    async def update_power_consumption(
        self,
        state_repository: StatesRepository,
        optimizer: Optimizer,
        grid_exported_power_data: DataBuffer,
    ) -> None:
        """Update the device based on the current pv availability."""
        if self._output_id is not None:
            state: bool = (
                self._output_state.value == "on" if self._output_state is not None else False
            )
            new_state = state
            if self.power_mode == PowerModes.PV:
                avg_300 = grid_exported_power_data.get_average_for(300)
                if avg_300 > self.nominal_power * (1 + POWER_HYSTERESIS):
                    new_state = True
                elif avg_300 < self.nominal_power * (1 - POWER_HYSTERESIS):
                    new_state = False
            elif self.power_mode == PowerModes.OPTIMIZED:
                power = optimizer.get_optimized_power(self._id)
                new_state = power > 0
            if state != new_state:
                state_repository.set_state(
                    StateId(
                        id=self._output_id,
                        channel=HOMEASSISTANT_CHANNEL,
                    ),
                    "on" if new_state else "off",
                )

    @property
    def consumed_energy(self) -> float:
        """The consumed energy of the device."""
        return self._consumed_energy.numeric_value if self._consumed_energy is not None else 0.0

    @property
    def icon(self) -> str:
        """The icon of the device."""
        return self._icon

    @property
    def power(self) -> float:
        """The current power used by the device."""
        return self._power.numeric_value if self._power else 0.0

    @property
    def available(self) -> bool:
        """Is the device available?."""
        return (
            self._consumed_energy is not None
            and self._consumed_energy.available
            and self._power is not None
            and self._power.available
        )

    @property
    def nominal_power(self) -> float:
        """The nominal power of the device."""
        return self._nominal_power if self._nominal_power is not None else DEFAULT_NOMINAL_POWER

    def restore_state(self, consumed_solar_energy: float, consumed_energy: float) -> None:
        """Restore a previously stored state."""
        super().restore_state(consumed_solar_energy, consumed_energy)
        self._consumed_energy = HomeassistantState("", str(consumed_energy))

    @property
    def state(self) -> str:
        """The state of the device. The state is `on` in case the device is running."""
        return self._state

    @property
    def has_state(self) -> bool:
        """Has this device a state."""
        return self._device_type is not None

    def get_load_info(self) -> LoadInfo | None:
        """Get the current deferrable load info."""
        if self._nominal_power is not None and self._nominal_duration is not None:
            if self.power_mode == PowerModes.OPTIMIZED:
                return LoadInfo(
                    device_id=self.id,
                    nominal_power=self._nominal_power,
                    duration=self._nominal_duration,
                    is_continous=False,
                    is_constant=self._is_constant if self._is_constant is not None else False,
                    is_deferrable=True,
                )

            if self.state == OnOffState.ON:
                return LoadInfo(
                    device_id=self.id,
                    nominal_power=self._nominal_power,
                    duration=self._nominal_duration - self.session_duration,
                    is_continous=False,
                    is_constant=self._is_constant if self._is_constant is not None else False,
                    is_deferrable=False,
                )

        return None



"""Mqtt connection for energy assistant."""

import logging
import random

import paho.mqtt.client as mqtt
from paho.mqtt.enums import CallbackAPIVersion

from energy_assistant.devices import State, StatesSingleRepository

MQTT_CHANNEL = "mqtt"


def on_message(client: mqtt.Client, userdata, message: mqtt.MQTTMessage) -> None:  # type: ignore
    """Handle received mqtt messages."""
    userdata.on_message_received(message.topic, str(message.payload.decode("utf-8")))


def on_connect(client: mqtt.Client, userdata, flags, rc, properties) -> None:  # type: ignore
    """Handle connecting to the mqtt server."""
    logging.info("Connected to mqtt with result code " + str(rc))
    userdata.subscribe_topics()


def on_disconnect(client: mqtt.Client, userdata, rc, properties) -> None:  # type: ignore
    """Handle disconnect from the mqtt server."""
    logging.info("Disconnected to mqtt with result code " + str(rc))
    # TODO: Handle reconnect


class MqttConnection(StatesSingleRepository):
    """Connecting to mqtt."""

    def __init__(self, host: str, username: str, password: str, topic: str) -> None:
        """Create an Mqtt Connection instance."""
        super().__init__(MQTT_CHANNEL)
        self._host = host
        self._username = username
        self._password = password
        self._topic = topic
        self._subscription_topics: set = set()
        self._client: mqtt.Client | None = None

    def connect(self) -> None:
        """Connect to the mqtt server."""
        try:
            self._client = mqtt.Client(
                CallbackAPIVersion.VERSION2,
                "energy_assistant" + str(random.randrange(1024)),
                userdata=self,
            )
            self._client.username_pw_set(self._username, self._password)
            self._client.will_set(f"{self._topic}/status", payload="offline", qos=0, retain=True)
            self._client.on_message = on_message
            self._client.on_connect = on_connect
            self._client.on_disconnect = on_disconnect
            self._client.connect(self._host)
            self._client.loop_start()
        except Exception:
            logging.exception("Error while connecting mqtt ")

    def add_subscription_topic(self, topic: str) -> None:
        """Add a subscription topic."""
        if topic not in self._subscription_topics:
            self._subscription_topics.add(topic)
            if self._client is not None:
                self._client.subscribe(topic)

    def subscribe_topics(self) -> None:
        """Subscribe to the registered topics on mqtt."""
        if self._client is not None:
            for topic in self._subscription_topics:
                self._client.subscribe(topic)

    def on_message_received(self, id: str, value: str) -> None:
        """Handle a received mqtt message."""
        self._read_states[id] = State(id, value)

    def read_states(self) -> None:
        """Read the states from the channel."""
        pass

    def write_states(self) -> None:
        """Write the states to the channel."""
        if self._client:
            for id, state in self._write_states.items():
                self._client.publish(id, state.value)



"""Use cases for home measurements."""

from datetime import date

from fastapi import HTTPException

from energy_assistant.api.device import OTHER_DEVICE
from energy_assistant.db import AsyncSession
from energy_assistant.models.home import HomeMeasurement

from .schema import (
    DeviceMeasurementPeriodSchema,
    HomeMeasurementDailySchema,
    HomeMeasurementDateSchema,
    HomeMeasurementPeriodSchema,
)


def add_others_device(
    home_measurement: HomeMeasurementPeriodSchema,
) -> HomeMeasurementPeriodSchema:
    """Add the date for the others device."""
    solar_consumed_energy = home_measurement.solar_consumed_energy
    consumed_energy = home_measurement.consumed_energy
    for device_measurement in home_measurement.device_measurements:
        solar_consumed_energy = solar_consumed_energy - device_measurement.solar_consumed_energy
        consumed_energy = consumed_energy - device_measurement.consumed_energy
    other_device = DeviceMeasurementPeriodSchema(
        solar_consumed_energy=solar_consumed_energy,
        consumed_energy=consumed_energy,
        device_id=OTHER_DEVICE,
    )
    home_measurement.device_measurements.append(other_device)
    return home_measurement


class ReadHomeMeasurementDifference:
    """Read the difference between two home measurementuse case."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a read home measurement use case."""
        self.async_session = session

    async def execute(self, from_date: date, to_date: date) -> HomeMeasurementPeriodSchema:
        """Execute the read home measurement use case."""
        async with self.async_session() as session:
            home_measurement_from = await HomeMeasurement.read_before_date(
                session, from_date, include_device_measurements=True
            )
            if not home_measurement_from:
                home_measurement_from = await HomeMeasurement.read_first(
                    session, include_device_measurements=True
                )
                if not home_measurement_from:
                    raise HTTPException(status_code=404)
            home_measurement_to = await HomeMeasurement.read_by_date(
                session, to_date, include_device_measurements=True
            )

            if not home_measurement_from:
                raise HTTPException(status_code=404)

            if not home_measurement_to:
                raise HTTPException(status_code=404)

            device_measurements = []
            for from_device in home_measurement_from.device_measurements:
                to_device = home_measurement_to.get_device_measurement(
                    from_device.device_id
                )  # home_measurement_to.device_measurements[index]
                if to_device is not None:
                    measurement = DeviceMeasurementPeriodSchema(
                        device_id=from_device.device_id,
                        solar_consumed_energy=to_device.solar_consumed_energy
                        - from_device.solar_consumed_energy,
                        consumed_energy=to_device.consumed_energy - from_device.consumed_energy,
                    )
                    device_measurements.append(measurement)

            result = HomeMeasurementPeriodSchema(
                consumed_energy=home_measurement_to.consumed_energy
                - home_measurement_from.consumed_energy,
                solar_consumed_energy=home_measurement_to.solar_consumed_energy
                - home_measurement_from.solar_consumed_energy,
                solar_produced_energy=home_measurement_to.solar_produced_energy
                - home_measurement_from.solar_produced_energy,
                grid_exported_energy=home_measurement_to.grid_exported_energy
                - home_measurement_from.grid_exported_energy,
                grid_imported_energy=home_measurement_to.grid_imported_energy
                - home_measurement_from.grid_imported_energy,
                device_measurements=device_measurements,
            )

            return add_others_device(result)


class ReadHomeMeasurementDaily:
    """Read the daily usage between two dates use case."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a read home measurement dates use case."""
        self.async_session = session

    async def execute(self, from_date: date, to_date: date) -> HomeMeasurementDailySchema:
        """Execute the read daily home measurement use case."""
        async with self.async_session() as session:
            last_measurement = await HomeMeasurement.read_before_date(
                session, from_date, include_device_measurements=True
            )
            if not last_measurement:
                last_measurement = await HomeMeasurement.read_first(
                    session, include_device_measurements=True
                )
                if not last_measurement:
                    raise HTTPException(status_code=404)
            result = []
            async for home_measurement in HomeMeasurement.read_between_dates(
                session, from_date, to_date, True
            ):
                device_measurements = []
                for from_device in last_measurement.device_measurements:
                    to_device = home_measurement.get_device_measurement(from_device.device_id)
                    if to_device is not None:
                        device_measurement = DeviceMeasurementPeriodSchema(
                            device_id=from_device.device_id,
                            solar_consumed_energy=to_device.solar_consumed_energy
                            - from_device.solar_consumed_energy,
                            consumed_energy=to_device.consumed_energy - from_device.consumed_energy,
                        )
                        device_measurements.append(device_measurement)
                measurement = HomeMeasurementDateSchema(
                    solar_consumed_energy=home_measurement.solar_consumed_energy
                    - last_measurement.solar_consumed_energy,
                    consumed_energy=home_measurement.consumed_energy
                    - last_measurement.consumed_energy,
                    solar_produced_energy=home_measurement.solar_produced_energy
                    - last_measurement.solar_produced_energy,
                    grid_imported_energy=home_measurement.grid_imported_energy
                    - last_measurement.grid_imported_energy,
                    grid_exported_energy=home_measurement.grid_exported_energy
                    - last_measurement.grid_exported_energy,
                    measurement_date=home_measurement.measurement_date,
                    device_measurements=device_measurements,
                )
                add_others_device(measurement)
                result.append(measurement)
                last_measurement = home_measurement
            return HomeMeasurementDailySchema(measurements=result)



"""Added power_mode

Revision ID: 81f94b4136ee
Revises: b6020d5b13fc
Create Date: 2023-07-28 19:22:18.654152

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "81f94b4136ee"
down_revision = "b6020d5b13fc"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    # op.alter_column('DeviceMeasurement', 'device_id', existing_type=sa.CHAR(length=32),  nullable=False)
    with op.batch_alter_table("DeviceMeasurement") as batch_op:
        batch_op.drop_column("name")
    op.add_column("devices", sa.Column("power_mode", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("devices", "power_mode")
    op.add_column("DeviceMeasurement", sa.Column("name", sa.VARCHAR(), nullable=False))
    # op.alter_column('DeviceMeasurement', 'device_id', existing_type=sa.CHAR(length=32), nullable=True)
    # ### end Alembic commands ###



"""Schemas for home measurement api."""

from pydantic import BaseModel


class ConfigModel(BaseModel):
    """Schema for the configuration."""

    config: dict


class ReadConfigResponse(ConfigModel):
    """API Response for reading the configuration."""

    pass



"""Config helper classes and functions."""

from energy_assistant.devices import Location


class DeviceConfigException(Exception):
    """Device configuration exception."""

    pass


def get_config_param(config: dict, param: str) -> str:
    """Get a config parameter as string or raise an exception if the parameter is not available."""
    result = config.get(param)
    if result is None:
        raise DeviceConfigException(f"Parameter {param} is missing in the configuration")
    else:
        return str(result)


def get_config_param_from_list(config: list, param: str) -> str | None:
    """Read config param from a list."""
    for item in config:
        value = item.get(param)
        if value is not None:
            return value
    return None


def get_float_param_from_list(config: list, param: str) -> float | None:
    """Read a float config param from a list."""
    for item in config:
        value = item.get(param)
        if value is not None:
            return float(value)
    return None


class EnergyAssistantConfig:
    """The energy assistant config."""

    def __init__(self, energy_assistant_config: dict, hass_config: dict) -> None:
        """Create a EnergyAssistantConfig instance."""
        self._config = {
            "energy_assistant": energy_assistant_config.copy(),
            "home_assistant": hass_config,
        }
        self._config["emhass"] = energy_assistant_config["emhass"].copy()
        del self._config["energy_assistant"]["emhass"]

    @property
    def config(self) -> dict:
        """Get the complete config."""
        return self._config

    @property
    def energy_assistant_config(self) -> dict:
        """Get the energy assistant config."""
        return self._config["energy_assistant"]

    @property
    def home_assistant_config(self) -> dict:
        """Get the home assistant config."""
        return self._config["home_assistant"]

    @property
    def emhass_config(self) -> dict:
        """Get the emhass config."""
        return self._config["emhass"]

    @property
    def location(self) -> Location:
        """Read the location from the Homeassistant configuration."""
        config = self.home_assistant_config

        return Location(
            latitude=config.get("latitude", ""),
            longitude=config.get("longitude", ""),
            elevation=config.get("elevation", ""),
            time_zone=config.get("time_zone", ""),
        )



"""Main module for the energy assistant application."""

import asyncio
import json
import logging
import os
import sys
import threading
from contextlib import asynccontextmanager, suppress
from datetime import date
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import AsyncIterator, Final

import alembic.config
import requests  # type: ignore
import yaml
from apscheduler.schedulers.asyncio import AsyncIOScheduler  # type: ignore
from colorlog import ColoredFormatter
from energy_assistant_frontend import where as locate_frontend
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from fastapi.staticfiles import StaticFiles
from sqlalchemy.ext.asyncio import AsyncSession

from energy_assistant.api.main import router as api_router
from energy_assistant.devices import StatesMultipleRepositories, StatesRepository
from energy_assistant.devices.config import EnergyAssistantConfig
from energy_assistant.devices.evcc import EvccDevice
from energy_assistant.devices.home import Home
from energy_assistant.devices.homeassistant import Homeassistant
from energy_assistant.devices.registry import DeviceTypeRegistry
from energy_assistant.EmhassOptimizer import EmhassOptimizer
from energy_assistant.mqtt import MqttConnection
from energy_assistant.settings import settings
from energy_assistant.storage import Database, get_async_session, session_storage
from energy_assistant.websocket import get_home_message, ws_manager, ws_router

from .constants import ROOT_LOGGER_NAME

FORMAT_DATE: Final = "%Y-%m-%d"
FORMAT_TIME: Final = "%H:%M:%S"
FORMAT_DATETIME: Final = f"{FORMAT_DATE} {FORMAT_TIME}"
MAX_LOG_FILESIZE = 1000000 * 10  # 10 MB


class EnergyAssistant:
    """Energy Assistant Application."""

    home: Home
    hass: Homeassistant | None
    optimizer: EmhassOptimizer
    mqtt: MqttConnection | None
    db: Database
    config: EnergyAssistantConfig
    should_stop = False


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncIterator:
    """Manage the startup and showdown."""
    ea = await init_app()
    app.energy_assistant = ea  # type: ignore
    bt = asyncio.create_task(background_task(ea))
    optimizer_task = asyncio.create_task(optimize(ea.optimizer))
    scheduler = AsyncIOScheduler()
    scheduler.add_job(
        daily_optimize, trigger="cron", args=[ea], hour="3", minute="0"
    )  # time is UTC
    scheduler.start()
    yield
    ea.should_stop = True
    await optimizer_task
    await bt
    scheduler.shutdown()
    print("Shutdown app")


app = FastAPI(title="energy-assistant", lifespan=lifespan)


origins = [
    "http://localhost",
    "http://localhost:5000",
    "http://localhost:5173",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(api_router, prefix="/api")
app.include_router(ws_router)


async def async_handle_state_update(
    ea: EnergyAssistant,
    state_repository: StatesRepository,
    session: AsyncSession,
) -> None:
    """Read the values from home assistant and process the update."""
    try:
        state_repository.read_states()

        await ea.home.update_state(state_repository)
        if ea.optimizer is not None:
            await ea.home.update_power_consumption(state_repository, ea.optimizer)
            ea.optimizer.update_repository_states(ea.home, state_repository)
        else:
            logging.error("The variable optimizer is None in async_handle_state_update")
        state_repository.write_states()
        if ea.db:
            if ea.home:
                await asyncio.gather(
                    ws_manager.broadcast(get_home_message(ea.home)),
                    ea.db.store_home_state(ea.home, session),
                )
            else:
                logging.error("The variable home is None in async_handle_state_update")
        else:
            logging.error("The variable db is None in async_handle_state_update")
        if ea.optimizer is not None and ea.home is not None:
            ea.optimizer.update_devices(ea.home)
    except Exception:
        logging.exception("error during sending refresh")


async def background_task(ea: EnergyAssistant) -> None:
    """Periodically read the values from home assistant and process the update."""
    last_update = date.today()
    async_session = await get_async_session()
    repositories: list[StatesRepository] = []
    if ea.hass is not None:
        repositories.append(ea.hass)
    if ea.mqtt is not None:
        repositories.append(ea.mqtt)
    state_repository = StatesMultipleRepositories(repositories)
    while not ea.should_stop:
        await asyncio.sleep(30)
        # delta_t = datetime.now().timestamp()
        # print("Start refresh from home assistant")
        today = date.today()
        try:
            if today != last_update:
                ea.home.store_energy_snapshot()
            last_update = today
            async with async_session() as session:
                await async_handle_state_update(ea, state_repository, session)
        except Exception:
            logging.exception("error in the background task")
        # print(f"refresh from home assistant completed in {datetime.now().timestamp() - delta_t} s")


def create_mqtt_connection(config: dict) -> MqttConnection | None:
    """Create an mqtt connection based on the config."""

    mqtt_config = config.get("mqtt")
    if mqtt_config is not None:
        mqtt_host = mqtt_config.get("host")
        mqtt_username = mqtt_config.get("username")
        mqtt_password = mqtt_config.get("password")
        mqtt_topic = mqtt_config.get("topic")
        mqtt_connection = MqttConnection(mqtt_host, mqtt_username, mqtt_password, mqtt_topic)
        mqtt_connection.connect()
        return mqtt_connection
    return None


def subscribe_mqtt_topics(mqtt_connection: MqttConnection, home: Home) -> None:
    """Subscribe the mqtt based devices on the mqtt connection."""

    for device in home.devices:
        if isinstance(device, EvccDevice):
            mqtt_connection.add_subscription_topic(device.evcc_mqtt_subscription_topic)


def create_hass_connection(config: dict) -> Homeassistant | None:
    """Create a connection to home assistant."""
    token: str | None = None
    url: str | None = None
    try:
        token = os.getenv("SUPERVISOR_TOKEN")
        if token is not None:
            logging.info(f"suvervisor token detected. len={len(token)}")
            url = "http://supervisor/core"

            headers = {
                "Authorization": f"Bearer {token}",
                "content-type": "application/json",
            }
            response = requests.get(f"{url}/api/states", headers=headers)
            logging.info(
                f"pinging homeassistant api succeeded. Status code = {response.status_code}"
            )
            if response.ok:
                logging.info(f"Using {url} to connect")
                hass = Homeassistant(url, token, False)
                return hass
    except Exception:
        logging.exception("Error while trying to connect to the homeassistant supervisor api")
        url = None
        token = None

    logging.info("Try to connect to home assistant based on the config file entries...")
    hass_config = config.get("homeassistant")
    if hass_config is not None:
        demo_mode = hass_config.get("demo_mode")
        url = hass_config.get("url")
        token = hass_config.get("token")
        if url is not None and token is not None:
            return Homeassistant(url, token, demo_mode)
    return None


def setup_logger(log_filename: str, level: str = "DEBUG") -> logging.Logger:
    """Initialize logger."""
    # define log formatter
    log_fmt = "%(asctime)s.%(msecs)03d %(levelname)s (%(threadName)s) [%(name)s] %(message)s"

    # base logging config for the root logger
    logging.basicConfig(level=logging.INFO)

    colorfmt = f"%(log_color)s{log_fmt}%(reset)s"
    logging.getLogger().handlers[0].setFormatter(
        ColoredFormatter(
            colorfmt,
            datefmt=FORMAT_DATETIME,
            reset=True,
            log_colors={
                "DEBUG": "cyan",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red",
            },
        )
    )

    # Capture warnings.warn(...) and friends messages in logs.
    # The standard destination for them is stderr, which may end up unnoticed.
    # This way they're where other messages are, and can be filtered as usual.
    logging.captureWarnings(True)

    # setup file handler
    # log_filename = os.path.join(data_path, "energy_assistant.log")
    file_handler = RotatingFileHandler(log_filename, maxBytes=MAX_LOG_FILESIZE, backupCount=1)
    # rotate log at each start
    with suppress(OSError):
        file_handler.doRollover()
    file_handler.setFormatter(logging.Formatter(log_fmt, datefmt=FORMAT_DATETIME))
    # file_handler.setLevel(logging.INFO)

    logger = logging.getLogger()
    logger.addHandler(file_handler)

    # apply the configured global log level to the (root) music assistant logger
    logging.getLogger(ROOT_LOGGER_NAME).setLevel(level)

    # silence some noisy loggers
    logging.getLogger("asyncio").setLevel(logging.WARNING)
    logging.getLogger("aiosqlite").setLevel(logging.WARNING)
    logging.getLogger("databases").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)

    sys.excepthook = lambda *args: logging.getLogger(None).exception(
        "Uncaught exception",
        exc_info=args,  # type: ignore[arg-type]
    )
    threading.excepthook = lambda args: logging.getLogger(None).exception(
        "Uncaught thread exception",
        exc_info=(  # type: ignore[arg-type]
            args.exc_type,
            args.exc_value,
            args.exc_traceback,
        ),
    )

    return logger


async def init_app() -> EnergyAssistant:
    """Initialize the application."""
    result = EnergyAssistant()

    hass_options_file = "/data/options.json"
    if os.path.isfile(hass_options_file):
        with open(hass_options_file, "rb") as _file:
            hass_options = json.loads(_file.read())
    else:
        hass_options = {}

    log_level = hass_options.get("log_level", "info").upper()

    logger = setup_logger(settings.LOG_FILE, log_level)

    config_file = hass_options.get("config_file", settings.CONFIG_FILE)

    logger.info("Hello from Energy Assistant")

    for option in hass_options:
        logging.info(f"{option}={hass_options[option]}")

    async_session = await get_async_session()
    db = Database()
    result.db = db

    device_registry_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "config/deviceregistry"
    )

    device_type_registry = DeviceTypeRegistry()
    device_type_registry.load(device_registry_path)

    logger.info(f"Loading config file {config_file}")
    try:
        with open(config_file) as stream:
            logger.debug(f"Successfully opened config file {config_file}")
            try:
                config = yaml.safe_load(stream)
                logger.debug(f"config file {config_file} successfully loaded")
            except yaml.YAMLError:
                logger.exception("Failed to parse the config file")
            except Exception:
                logger.exception("Failed to parse the config file")
            else:
                hass = create_hass_connection(config)
                result.hass = hass
                result.config = EnergyAssistantConfig(
                    config, hass.get_config() if hass is not None else {}
                )
                if hass is not None:
                    hass.read_states()
                    optimizer = EmhassOptimizer(settings.DATA_FOLDER, result.config, hass)
                    result.optimizer = optimizer
                    app.optimizer = optimizer  # type: ignore

                mqtt_connection: MqttConnection | None = create_mqtt_connection(config)
                result.mqtt = mqtt_connection

                home_config = config.get("home")
                if home_config is not None and home_config.get("name") is not None:
                    home = Home(home_config, session_storage, device_type_registry)
                    result.home = home
                    app.home = home  # type: ignore
                    if mqtt_connection is not None:
                        subscribe_mqtt_topics(mqtt_connection, home)

                    async with async_session() as session:
                        await db.update_devices(
                            home, session, session_storage, device_type_registry
                        )

                        await db.restore_home_state(home, session)
                else:
                    logger.error(f"home not found in config file: {config}")
                logger.info("Initialization completed")
    except Exception:
        logger.exception("Initialization of the app failed")
    return result


async def optimize(optimizer: EmhassOptimizer) -> None:
    """Optimize the forecast."""
    try:
        optimizer.forecast_model_fit(True)
    except Exception:
        logging.exception(
            "Optimization of the power consumption forecast model failed, probably due to missing history data in Home Assistant."
        )


def daily_optimize(ea: EnergyAssistant) -> None:
    """Optimize once a day."""
    try:
        optimizer = ea.optimizer
        if optimizer is not None:
            logging.info("Start optimizer run")
            optimizer.dayahead_forecast_optim()
    except Exception:
        logging.exception("Daily optimization run failed")


@app.get("/check", include_in_schema=False)
async def health() -> JSONResponse:
    """Test the web server with a ping."""
    return JSONResponse({"message": "It worked!!"})


# This needs to be the last mount
frontend_dir = locate_frontend()
app.mount("/", StaticFiles(directory=frontend_dir, html=True), name="frontend")


def main() -> None:
    """Start energy assistant."""
    import uvicorn

    try:
        alembic_config = Path(__file__).parent / "alembic.ini"

        alembicArgs = [
            "-c",
            str(alembic_config),
            "--raiseerr",
            "upgrade",
            "head",
        ]
        alembic.config.main(argv=alembicArgs)  # type: ignore
    except Exception:
        print("Alembic Migration failed")
        logging.exception("Alembic Migration failed")

    uvicorn.run(app, host="0.0.0.0", port=5000)


if __name__ == "__main__":
    main()



"""Schemas for home measurement api."""

import uuid
from datetime import date

from pydantic import BaseModel, ConfigDict


class DeviceMeasurementPeriodSchema(BaseModel):
    """Schema class for a device measurement."""

    device_id: uuid.UUID

    solar_consumed_energy: float
    consumed_energy: float

    model_config = ConfigDict(from_attributes=True)


class HomeMeasurementPeriodSchema(BaseModel):
    """Schema class for a home measurement."""

    solar_consumed_energy: float
    consumed_energy: float
    solar_produced_energy: float
    grid_imported_energy: float
    grid_exported_energy: float

    device_measurements: list[DeviceMeasurementPeriodSchema]

    model_config = ConfigDict(from_attributes=True)


class ReadHomeMeasurementDifferenceResponse(HomeMeasurementPeriodSchema):
    """API Response for reading home measurements."""

    pass


class HomeMeasurementDateSchema(HomeMeasurementPeriodSchema):
    """Schema class for a home measurement summary of a specific day."""

    measurement_date: date

    model_config = ConfigDict(from_attributes=True)


class HomeMeasurementDailySchema(BaseModel):
    """Schema class for daily home measurements."""

    measurements: list[HomeMeasurementDateSchema]

    model_config = ConfigDict(from_attributes=True)


class HomeMeasurementDailyResponse(HomeMeasurementDailySchema):
    """Schema for the daily home measurements response."""

    pass



"""Views for home measurement API."""

from datetime import date

from fastapi import APIRouter, Depends, Path, Request

from energy_assistant.models.schema import HomeMeasurementSchema

from .schema import ReadAllHomeMeasurementResponse, ReadHomeMeasurementResponse
from .use_cases import (
    DeleteHomeMeasurement,
    ReadAllHomeMeasurement,
    ReadHomeMeasurement,
    ReadHomeMeasurementByDate,
    ReadHomeMeasurementLastBeforeDate,
)

router = APIRouter(prefix="/homemeasurements")


@router.get("", response_model=ReadAllHomeMeasurementResponse)
async def read_all(
    request: Request, use_case: ReadAllHomeMeasurement = Depends(ReadAllHomeMeasurement)
) -> ReadAllHomeMeasurementResponse:
    """Rest end point for read all home measurements."""
    return ReadAllHomeMeasurementResponse(
        home_measurements=[home_measurement async for home_measurement in use_case.execute()]
    )


@router.get(
    "/{HomeMeasurement_id}",
    response_model=ReadHomeMeasurementResponse,
)
async def read(
    request: Request,
    home_measurement_id: int = Path(..., description=""),
    use_case: ReadHomeMeasurement = Depends(ReadHomeMeasurement),
) -> HomeMeasurementSchema:
    """REST end point for read a home measurement."""
    return await use_case.execute(home_measurement_id)


@router.get(
    "/by_date/{measurement_date}",
    response_model=ReadHomeMeasurementResponse,
)
async def read_by_date(
    request: Request,
    measurement_date: date = Path(..., description=""),
    use_case: ReadHomeMeasurementByDate = Depends(ReadHomeMeasurementByDate),
) -> HomeMeasurementSchema:
    """REST end point for read a home measurement by date."""
    return await use_case.execute(measurement_date)


@router.get(
    "/before_date/{measurement_date}",
    response_model=ReadHomeMeasurementResponse,
)
async def read_before_date(
    request: Request,
    measurement_date: date = Path(..., description=""),
    use_case: ReadHomeMeasurementLastBeforeDate = Depends(ReadHomeMeasurementLastBeforeDate),
) -> HomeMeasurementSchema:
    """REST end point for read the last home measurement before a date."""
    return await use_case.execute(measurement_date)


@router.delete("/{HomeMeasurement_id}", status_code=204)
async def delete(
    request: Request,
    home_measurement_id: int = Path(..., description=""),
    use_case: DeleteHomeMeasurement = Depends(DeleteHomeMeasurement),
) -> None:
    """REST end point for delete a home measurement."""
    await use_case.execute(home_measurement_id)



"""The settings for the Energy Assistant-."""

import os
from pathlib import Path

from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """Settings for the Energy Assistant."""

    DB_URI: str
    ECHO_SQL: bool
    CONFIG_FILE: str
    LOG_FILE: str
    DATA_FOLDER: str

    model_config = SettingsConfigDict(
        env_file=Path(__file__).parent / f"config/{os.environ['APP_CONFIG_FILE']}.env",
        case_sensitive=True,
    )


settings = Settings.model_validate({})



"""Forecast data models."""

from datetime import datetime

from pydantic import BaseModel


class ForecastSerieSchema(BaseModel):
    """Schema for a forecast data series."""

    name: str
    data: list[float]


class ForecastSchema(BaseModel):
    """Schema for the forecast."""

    cost: float
    solar_energy: float
    consumed_energy: float
    time: list[datetime]
    series: list[ForecastSerieSchema]



"""Tests for utility meters."""

from energy_assistant.devices.utility_meter import UtilityMeter


def test_utility_meter() -> None:
    """Test the utility meter."""
    meter = UtilityMeter("energy")
    assert meter.energy == 0
    meter.update_energy(10)
    assert meter.energy == 10
    meter.update_energy(1)
    assert meter.energy == 10
    meter.update_energy(5)
    assert meter.energy == 14



"""State value supporting the different representation of the states like single value, templates."""

import logging

from jinja2 import Environment, UndefinedError

from energy_assistant.constants import ROOT_LOGGER_NAME
from energy_assistant.devices import State, StatesRepository

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)
environment = Environment()


class CalculatedState(State):
    """A numeric, calculated state."""

    def __init__(self, value: float | str | None) -> None:
        """Create a calculated state instance."""
        if value is None:
            super().__init__("calculated", "0")
        elif isinstance(value, float):
            super().__init__("calculated", str(value))
        else:
            super().__init__("calculated", value)
        self._available = value is not None


class StateValue:
    """State value supporting the different representation of the states like single value, templates."""

    def __init__(self, config: dict | str) -> None:
        """Create a state value instance."""
        self._value_id: str | None = None
        self._template = None
        self._scale: float = 1.0
        if isinstance(config, str):
            self._value_id = config
        else:
            self._value_id = config.get("value")
            self._scale = config.get("scale", 1)
            inverted: bool | None = config.get("inverted")
            if inverted is not None and inverted:
                self._scale = -self._scale
            template: str | None = config.get("template")

            if template is not None:
                self._template = environment.from_string(template)

    def evaluate(self, state_repository: StatesRepository) -> State:
        """Evaluate the value."""
        result: State | None = None
        if self._value_id is not None:
            result = state_repository.get_state(self._value_id)
        elif self._template is not None:
            try:
                value = self._template.render(state_repository.get_template_states())
                result = CalculatedState(value)
            except UndefinedError as error:
                LOGGER.warning(f"undefined variable in expression: {error}")
                return CalculatedState(None)
        if result is not None:
            return CalculatedState(result.numeric_value * self._scale)
        return CalculatedState(result)

    def set_scale(self, scale: float) -> None:
        """Set the scale for the value. Evaluate multiplies the result with the scale."""
        self._scale = scale

    def invert_value(self) -> None:
        """Set the value to inverted. Evaluate multiples the result with -1."""
        self._scale = -self._scale



"""Use cases for devices."""

import logging

from energy_assistant.api.config.schema import ConfigModel
from energy_assistant.constants import ROOT_LOGGER_NAME
from energy_assistant.db import AsyncSession
from energy_assistant.devices.config import EnergyAssistantConfig

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)


class ReadConfiguration:
    """Read the configuration use case."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a read configuration use case."""
        self.async_session = session

    async def execute(self, config: EnergyAssistantConfig) -> ConfigModel:
        """Execute the read configuration use case."""
        result = ConfigModel.model_validate({"config": config.config})
        return result



"""Tests for the main module."""

import pytest
from httpx import AsyncClient


@pytest.mark.asyncio
async def test_health(ac: AsyncClient) -> None:
    """Test the healt check REST endpoint."""
    # client = await anext(ac)
    response = await ac.get("/check")
    assert response.status_code == 200



"""Tests for state value class."""

from energy_assistant.devices import State, StatesSingleRepository
from energy_assistant.devices.state_value import StateValue


class StatesSingleRepositoryMock(StatesSingleRepository):
    """Mock class for a single state repository."""

    def __init__(self, read_states: dict[str, State]) -> None:
        """Create an instance of the StatesSingleRepositoryMock class."""
        super().__init__("test")
        self._read_states = read_states

    def read_states(self) -> None:
        """Read the states from the channel."""
        pass

    def write_states(self) -> None:
        """Write the states to the channel."""
        pass


def test_state_value() -> None:
    """Test the state value."""
    state_repository_read = {
        "sensor.power": State("sensor.power", "10.1"),
        "sensor.energy_low": State("sensor.energy_low", "856"),
        "sensor.energy_high": State("sensor.energy_high", "7"),
    }
    state_repository = StatesSingleRepositoryMock(state_repository_read)
    template_states = state_repository.get_template_states()
    assert "sensor" in template_states
    assert len(template_states["sensor"]) == 3

    state_value = StateValue({"template": "{{sensor.energy_low + sensor.energy_high * 1000}}"})
    assert state_value.evaluate(state_repository).numeric_value == 7856

    state_value = StateValue("sensor.power")
    assert state_value.evaluate(state_repository).numeric_value == 10.1

    state_value = StateValue("sensor.power_unknown")
    assert state_value.evaluate(state_repository).available is False

    state_value = StateValue({"template": "{{sensor.energy_unknown + sensor.energy_high * 1000}}"})
    assert state_value.evaluate(state_repository).available is False

    state_value = StateValue("sensor.power")
    state_value.set_scale(0.001)
    assert state_value.evaluate(state_repository).numeric_value == 0.0101

    state_value = StateValue({"value": "sensor.power"})
    assert state_value.evaluate(state_repository).numeric_value == 10.1

    state_value = StateValue({"value": "sensor.power", "inverted": True})
    assert state_value.evaluate(state_repository).numeric_value == -10.1

    state_value = StateValue(
        {"template": "{{sensor.energy_low + sensor.energy_high * 1000}}", "scale": 0.001}
    )
    assert state_value.evaluate(state_repository).numeric_value == 7.856

    state_value = StateValue({"value": "sensor.power", "scale": 0.001})
    assert state_value.evaluate(state_repository).numeric_value == 0.0101



"""Views for home measurement API."""

from fastapi import APIRouter, Depends, HTTPException, Request

from energy_assistant.models.forecast import ForecastSchema

from .schema import CreateModelResponse, TuneModelResponse
from .use_cases import CreateModel, ReadForecast, TuneModel

router = APIRouter(prefix="/forecast")


@router.get("", response_model=ForecastSchema)
async def read_all(
    request: Request, use_case: ReadForecast = Depends(ReadForecast)
) -> ForecastSchema:
    """Rest end point for read all devices."""
    energy_assistant = (
        request.app.energy_assistant if hasattr(request.app, "energy_assistant") else None
    )
    if energy_assistant is None:
        raise HTTPException(status_code=500)
    return await use_case.execute(energy_assistant.optimizer)


@router.post("/create_model", response_model=CreateModelResponse)
async def create_model(
    request: Request, days_to_retrieve: int, use_case: CreateModel = Depends(CreateModel)
) -> CreateModelResponse:
    """Create the machine learning forecast model."""
    energy_assistant = (
        request.app.energy_assistant if hasattr(request.app, "energy_assistant") else None
    )
    if energy_assistant is None:
        raise HTTPException(status_code=500)
    return await use_case.execute(days_to_retrieve, energy_assistant.optimizer)


@router.post("/tune_model", response_model=TuneModelResponse)
async def tune_model(
    request: Request, use_case: TuneModel = Depends(TuneModel)
) -> TuneModelResponse:
    """Tune the machine learning forecast model."""
    energy_assistant = (
        request.app.energy_assistant if hasattr(request.app, "energy_assistant") else None
    )
    if energy_assistant is None:
        raise HTTPException(status_code=500)
    return await use_case.execute(energy_assistant.optimizer)



"""Mqtt connection for energy assistant."""

import copy
import json
import logging
import pathlib
import pickle
import uuid
from datetime import datetime, timezone
from typing import Tuple

import numpy as np
import pandas as pd
from emhass import utils  # type: ignore
from emhass.forecast import Forecast  # type: ignore
from emhass.machine_learning_forecaster import MLForecaster  # type: ignore
from emhass.optimization import Optimization  # type: ignore
from emhass.retrieve_hass import RetrieveHass  # type: ignore
from sklearn.metrics import r2_score  # type: ignore

from energy_assistant import Optimizer
from energy_assistant.devices import LoadInfo, Location, StateId, StatesRepository
from energy_assistant.devices.analysis import DataBuffer, create_timeseries_from_const
from energy_assistant.devices.config import EnergyAssistantConfig
from energy_assistant.devices.home import Home
from energy_assistant.devices.homeassistant import HOMEASSISTANT_CHANNEL, Homeassistant
from energy_assistant.models.forecast import ForecastSchema, ForecastSerieSchema

from .constants import ROOT_LOGGER_NAME

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)

SENSOR_POWER_NO_VAR_LOADS = "power_load_no_var_loads"
DEFAULT_HASS_ENTITY_PREFIX = "em"
DEFAULT_COST_FUNC = "profit"
LOAD_FORECAST_MODEL_TYPE = "load_forecast"


class EmhassOptimizer(Optimizer):
    """Optimizer based on Emhass."""

    def __init__(
        self, data_folder: str, config: EnergyAssistantConfig, hass: Homeassistant
    ) -> None:
        """Create an emhass optimizer instance."""
        self._data_folder: pathlib.Path = pathlib.Path(data_folder)
        self._logger = LOGGER
        self._hass_url: str = hass.url
        if self._hass_url is not None and self._hass_url[-1] != "/":
            self._hass_url = self._hass_url + "/"
        self._hass_token: str = hass.token
        self._location: Location = config.location

        home_config = config.energy_assistant_config.get("home")
        self._solar_power_id: str | None = None
        if home_config is not None:
            self._solar_power_id = home_config.get("solar_power")
        self._emhass_config: dict | None = config.emhass_config
        self._power_no_var_loads_id = (
            f"sensor.{DEFAULT_HASS_ENTITY_PREFIX}_{SENSOR_POWER_NO_VAR_LOADS}"
        )
        if self._emhass_config is not None:
            self._cost_fun = self._emhass_config.get("costfun", DEFAULT_COST_FUNC)
            self._hass_entity_prefix = self._emhass_config.get(
                "hass_entity_prefix", DEFAULT_HASS_ENTITY_PREFIX
            )
            self._power_no_var_loads_id = (
                f"sensor.{self._hass_entity_prefix}_{SENSOR_POWER_NO_VAR_LOADS}"
            )
            params = json.dumps(self._emhass_config)
            RetrieveHass_conf, optim_conf, plant_conf = utils.get_yaml_parse(
                pathlib.Path(), False, params=params
            )
            # Patch variables with Energy Assistant Config
            RetrieveHass_conf["hass_url"] = self._hass_url
            RetrieveHass_conf["long_lived_token"] = self._hass_token
            RetrieveHass_conf["var_PV"] = self._solar_power_id
            if "var_load" not in RetrieveHass_conf:
                RetrieveHass_conf["var_load"] = self._power_no_var_loads_id
            else:
                self._power_no_var_loads_id = RetrieveHass_conf["var_load"]
            RetrieveHass_conf["var_replace_zero"] = [self._solar_power_id]
            RetrieveHass_conf["var_interp"] = [
                self._solar_power_id,
                self._power_no_var_loads_id,
            ]

            RetrieveHass_conf["time_zone"] = self._location.get_time_zone()
            RetrieveHass_conf["lat"] = self._location.latitude
            RetrieveHass_conf["lon"] = self._location.longitude
            RetrieveHass_conf["alt"] = self._location.elevation

            optim_conf["num_def_loads"] = 0

            self._RetrieveHass_conf = RetrieveHass_conf
            self._optim_conf = optim_conf
            self._plant_conf = plant_conf

            self._method_ts_round = RetrieveHass_conf.get("method_ts_round")

            # Define main objects
            self._RetrieveHass = RetrieveHass(
                self._hass_url,
                self._hass_token,
                RetrieveHass_conf["freq"],
                self._location.get_time_zone(),
                params,
                self._data_folder,
                self._logger,
                get_data_from_file=False,
            )  # type: ignore

        if self._cost_fun is None:
            self._cost_fun = "profit"
        if self._method_ts_round is None:
            self._method_ts_round = "nearest"

        self._day_ahead_forecast: pd.DataFrame | None = None
        self._optimzed_devices: list = []
        self._projected_load_devices: list[LoadInfo] = []
        self._pv: DataBuffer = DataBuffer()
        self._no_var_loads: DataBuffer = DataBuffer()

    def update_repository_states(self, home: Home, state_repository: StatesRepository) -> None:
        """Calculate the power of the non variable/non controllable loads."""
        power = home.home_consumption_power
        for device in home.devices:
            if device.power_controllable:
                power = power - device.power
        if power < 0:
            power = 0.0
        self._no_var_loads.add_data_point(power)
        attributes = {
            "unit_of_measurement": "W",
            "state_class": "measurement",
            "device_class": "power",
        }
        state_repository.set_state(
            StateId(id=self._power_no_var_loads_id, channel=HOMEASSISTANT_CHANNEL),
            str(power),
            attributes,
        )
        state_repository.set_state(
            StateId(id=f"sensor.{self._hass_entity_prefix}_p_pv", channel=HOMEASSISTANT_CHANNEL),
            str(self._get_forecast_value("P_PV")),
            attributes,
        )
        state_repository.set_state(
            StateId(
                id=f"sensor.{self._hass_entity_prefix}_p_consumption", channel=HOMEASSISTANT_CHANNEL
            ),
            str(self._get_forecast_value("P_Load")),
            attributes,
        )

    def perfect_forecast_optim(
        self, save_data_to_file: bool = True, debug: bool = False
    ) -> pd.DataFrame:
        """Perform a call to the perfect forecast optimization routine.

        :param input_data_dict:  A dictionary with multiple data used by the action functions
        :type input_data_dict: dict
        :param logger: The passed logger object
        :type logger: logging object
        :param save_data_to_file: Save optimization results to CSV file
        :type save_data_to_file: bool, optional
        :param debug: A debug option useful for unittests
        :type debug: bool, optional
        :return: The output data of the optimization
        :rtype: pd.DataFrame

        """
        self._logger.info("Setting up needed data")

        # Treat runtimeparams
        params: str = ""
        params, RetrieveHass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            None,
            json.dumps(self._emhass_config),
            self._RetrieveHass_conf,
            self._optim_conf,
            self._plant_conf,
            "perfect-optim",
            self._logger,
        )  # type: ignore
        fcst = Forecast(
            self._RetrieveHass_conf,
            self._optim_conf,
            self._plant_conf,
            params,
            str(self._data_folder),
            self._logger,
            get_data_from_file=False,
        )
        opt = Optimization(
            self._RetrieveHass_conf,
            self._optim_conf,
            self._plant_conf,
            fcst.var_load_cost,
            fcst.var_prod_price,
            self._cost_fun,
            str(self._data_folder),
            self._logger,
        )

        days_list = utils.get_days_list(self._RetrieveHass_conf["days_to_retrieve"])
        var_list = [self._solar_power_id, self._power_no_var_loads_id]
        self._RetrieveHass.get_data(
            days_list, var_list, minimal_response=False, significant_changes_only=False
        )
        self._RetrieveHass.prepare_data(
            self._RetrieveHass_conf["var_load"],
            load_negative=self._RetrieveHass_conf["load_negative"],
            set_zero_min=self._RetrieveHass_conf["set_zero_min"],
            var_replace_zero=self._RetrieveHass_conf["var_replace_zero"],
            var_interp=self._RetrieveHass_conf["var_interp"],
        )
        df_input_data = self._RetrieveHass.df_final.copy()

        self._logger.info("Performing perfect forecast optimization")
        # Load cost and prod price forecast
        df_input_data = fcst.get_load_cost_forecast(
            df_input_data, method=fcst.optim_conf["load_cost_forecast_method"]
        )
        df_input_data = fcst.get_prod_price_forecast(
            df_input_data, method=fcst.optim_conf["prod_price_forecast_method"]
        )
        opt_res = opt.perform_perfect_forecast_optim(df_input_data, days_list)
        # Save CSV file for analysis
        if save_data_to_file:
            filename = f"opt_res_perfect_optim_{self._cost_fun}.csv"
        else:  # Just save the latest optimization results
            filename = "opt_res_perfect_optim_latest.csv"
        if not debug:
            opt_res.to_csv(self._data_folder / filename, index_label="timestamp")
        return opt_res

    def get_ml_runtime_params(self) -> dict:
        """Get the emhass runtime params for the machine learning load prediction."""
        freq = self._RetrieveHass_conf["freq"].total_seconds() / 3600

        runtimeparams: dict = {
            "num_def_loads": len(self._optimzed_devices),
            "P_deferrable_nom": [device.nominal_power for device in self._optimzed_devices],
            "def_total_hours": [
                max(round(device.duration / 3600), 1) for device in self._optimzed_devices
            ],
            "def_start_timestep": [device.start_timestep for device in self._optimzed_devices],
            "def_end_timestep": [device.end_timestep for device in self._optimzed_devices],
            "treat_def_as_semi_cont": [
                not device.is_continous for device in self._optimzed_devices
            ],
            "set_def_constant": [device.is_constant for device in self._optimzed_devices],
            "days_to_retrieve": self._RetrieveHass_conf.get("days_to_retrieve", 10),
            "model_type": LOAD_FORECAST_MODEL_TYPE,
            "var_model": self._power_no_var_loads_id,
            "sklearn_model": "KNeighborsRegressor",
            "num_lags": int(24 / freq),  # should be one day * 30 min
            "split_date_delta": "48h",
            "perform_backtest": True,
        }
        return runtimeparams

    def dayahead_forecast_optim(self, save_data_to_file: bool = False, debug: bool = False) -> None:
        """Perform a call to the day-ahead optimization routine.

        :param save_data_to_file: Save optimization results to CSV file
        :type save_data_to_file: bool, optional
        :param debug: A debug option useful for unittests
        :type debug: bool, optional

        """
        self._logger.info("Setting up needed data for a day ahead forecast")

        # Fall back to the naive method in case the model for the mlforecaster does not exist
        filename_path = self._data_folder / "load_forecast_mlf.pkl"
        if not filename_path.exists():
            self._optim_conf["load_forecast_method"] = "naive"
            self._logger.warning("Falling back to the naive load forecaster.")

        # Treat runtimeparams
        params: str = ""
        params, RetrieveHass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            json.dumps(self.get_ml_runtime_params()),
            json.dumps(self._emhass_config),
            self._RetrieveHass_conf,
            self._optim_conf,
            self._plant_conf,
            "dayahead-optim",
            self._logger,
        )  # type: ignore
        fcst = Forecast(
            self._RetrieveHass_conf,
            self._optim_conf,
            self._plant_conf,
            params,
            str(self._data_folder),
            self._logger,
            get_data_from_file=False,
        )
        opt = Optimization(
            self._RetrieveHass_conf,
            self._optim_conf,
            self._plant_conf,
            fcst.var_load_cost,
            fcst.var_prod_price,
            self._cost_fun,
            str(self._data_folder),
            self._logger,
        )

        df_weather = fcst.get_weather_forecast(method=self._optim_conf["weather_forecast_method"])
        P_PV_forecast = fcst.get_power_from_weather(df_weather)
        try:
            P_load_forecast = fcst.get_load_forecast(
                method=self._optim_conf["load_forecast_method"]
            )
            P_load_forecast_values = np.array(P_load_forecast.values)
        except Exception:
            self._logger.warning(
                "Forecasting the load failed, probably due to missing history data in Home Assistant. "
            )
            avg_non_var_power = self._no_var_loads.average()
            P_load_forecast = pd.Series(
                [avg_non_var_power for x in P_PV_forecast.values],
                index=P_PV_forecast.index,
            )
            P_load_forecast_values = np.array([avg_non_var_power for x in P_PV_forecast.values])

        freq = self._RetrieveHass_conf["freq"]

        df_input_data_dayahead = pd.DataFrame(
            np.transpose(np.vstack([np.array(P_PV_forecast.values), P_load_forecast_values])),
            index=P_PV_forecast.index,
            columns=["P_PV_forecast", "P_non_deferrable_load_forecast"],
        )

        projected_load = None
        for load_info in self._projected_load_devices:
            if not load_info.is_deferrable:
                series = create_timeseries_from_const(
                    load_info.nominal_power, pd.Timedelta(int(load_info.duration), "s"), freq
                )
                if projected_load is None:
                    projected_load = series
                else:
                    projected_load = projected_load + series

        if projected_load is not None:
            projected_load = projected_load.tz_convert(self._location.get_time_zone())
            df_input_data_dayahead["P_projected_load"] = projected_load
            df_input_data_dayahead["P_projected_load"] = df_input_data_dayahead[
                "P_projected_load"
            ].fillna(0)
        else:
            df_input_data_dayahead["P_projected_load"] = 0

        df_input_data_dayahead["P_load_forecast"] = (
            df_input_data_dayahead["P_projected_load"]
            + df_input_data_dayahead["P_non_deferrable_load_forecast"]
        )
        P_load_forecast = df_input_data_dayahead["P_load_forecast"]
        df_input_data_dayahead = utils.set_df_index_freq(df_input_data_dayahead)

        # params_dayahead: dict = json.loads(params)
        # if 'prediction_horizon' in params_dayahead['passed_data'] and params_dayahead['passed_data']['prediction_horizon'] is not None:
        #    prediction_horizon = params_dayahead['passed_data']['prediction_horizon']
        #    df_input_data_dayahead = copy.deepcopy(df_input_data_dayahead)[df_input_data_dayahead.index[0]:df_input_data_dayahead.index[prediction_horizon-1]]

        self._logger.info("Performing day-ahead forecast optimization")
        # Load cost and prod price forecast
        df_input_data_dayahead = fcst.get_load_cost_forecast(
            df_input_data_dayahead, method=fcst.optim_conf["load_cost_forecast_method"]
        )
        df_input_data_dayahead = fcst.get_prod_price_forecast(
            df_input_data_dayahead, method=fcst.optim_conf["prod_price_forecast_method"]
        )
        self._day_ahead_forecast = opt.perform_dayahead_forecast_optim(
            df_input_data_dayahead, P_PV_forecast, P_load_forecast
        )
        if self._day_ahead_forecast is not None:
            self._day_ahead_forecast["P_projected_load"] = df_input_data_dayahead[
                "P_projected_load"
            ].copy()

        if not debug and self._day_ahead_forecast is not None:
            # Save CSV file for publish_data
            if save_data_to_file:
                today = datetime.now(timezone.utc).replace(
                    hour=0, minute=0, second=0, microsecond=0
                )
                filename = "opt_res_dayahead_" + today.strftime("%Y_%m_%d") + ".csv"
            else:  # Just save the latest optimization results
                filename = "opt_res_latest.csv"
            self._day_ahead_forecast.to_csv(self._data_folder / filename, index_label="timestamp")

    def naive_mpc_optim(self, save_data_to_file: bool = False, debug: bool = False) -> pd.DataFrame:
        """Perform a call to the naive Model Predictive Controller optimization routine.

        :param input_data_dict:  A dictionary with multiple data used by the action functions
        :type input_data_dict: dict
        :param logger: The passed logger object
        :type logger: logging object
        :param save_data_to_file: Save optimization results to CSV file
        :type save_data_to_file: bool, optional
        :param debug: A debug option useful for unittests
        :type debug: bool, optional
        :return: The output data of the optimization
        :rtype: pd.DataFrame

        """

        self._logger.info("Setting up needed data")

        runtimeparams: dict = {
            "prediction_horizon": 48  # How many 30 min slots do we predict -> 24h
        }

        # Treat runtimeparams
        params: str = ""
        params, RetrieveHass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            json.dumps(runtimeparams),
            json.dumps(self._emhass_config),
            self._RetrieveHass_conf,
            self._optim_conf,
            self._plant_conf,
            "naive-mpc-optim",
            self._logger,
        )  # type: ignore
        fcst = Forecast(
            self._RetrieveHass_conf,
            self._optim_conf,
            self._plant_conf,
            params,
            str(self._data_folder),
            self._logger,
            get_data_from_file=False,
        )
        opt = Optimization(
            self._RetrieveHass_conf,
            self._optim_conf,
            self._plant_conf,
            fcst.var_load_cost,
            fcst.var_prod_price,
            self._cost_fun,
            str(self._data_folder),
            self._logger,
        )

        # Retrieve data from hass
        days_list = utils.get_days_list(1)
        var_list = [self._solar_power_id, self._power_no_var_loads_id]
        self._RetrieveHass.get_data(
            days_list, var_list, minimal_response=False, significant_changes_only=False
        )
        self._RetrieveHass.prepare_data(
            self._RetrieveHass_conf["var_load"],
            load_negative=self._RetrieveHass_conf["load_negative"],
            set_zero_min=self._RetrieveHass_conf["set_zero_min"],
            var_replace_zero=self._RetrieveHass_conf["var_replace_zero"],
            var_interp=self._RetrieveHass_conf["var_interp"],
        )
        df_input_data = self._RetrieveHass.df_final.copy()

        # Get PV and load forecasts
        df_weather = fcst.get_weather_forecast(method=self._optim_conf["weather_forecast_method"])
        P_PV_forecast = fcst.get_power_from_weather(
            df_weather, set_mix_forecast=True, df_now=df_input_data
        )
        P_load_forecast = fcst.get_load_forecast(
            method=self._optim_conf["load_forecast_method"],
            set_mix_forecast=True,
            df_now=df_input_data,
        )
        df_input_data_dayahead = pd.concat(
            [
                pd.Series(P_PV_forecast, name="P_PV_forecast"),
                pd.Series(P_load_forecast, name="P_load_forecast"),
            ],
            axis=1,
        )
        df_input_data_dayahead = utils.set_df_index_freq(df_input_data_dayahead)

        # params_naive_mpc_optim: dict = json.loads(params)
        # if 'prediction_horizon' in params_naive_mpc_optim['passed_data'] and params_naive_mpc_optim['passed_data']['prediction_horizon'] is not None:
        #    prediction_horizon = params_naive_mpc_optim['passed_data']['prediction_horizon']
        #    df_input_data_dayahead = copy.deepcopy(df_input_data_dayahead)[df_input_data_dayahead.index[0]:df_input_data_dayahead.index[prediction_horizon-1]]

        self._logger.info("Performing naive MPC optimization")
        # Load cost and prod price forecast
        df_input_data_dayahead = fcst.get_load_cost_forecast(
            df_input_data_dayahead, method=fcst.optim_conf["load_cost_forecast_method"]
        )
        df_input_data_dayahead = fcst.get_prod_price_forecast(
            df_input_data_dayahead, method=fcst.optim_conf["prod_price_forecast_method"]
        )

        # The specifics params for the MPC at runtime
        # TODO: Make this real parameters
        params_dict = json.loads(params)
        prediction_horizon = params_dict["passed_data"]["prediction_horizon"]
        soc_init = params_dict["passed_data"]["soc_init"]
        soc_final = params_dict["passed_data"]["soc_final"]
        def_total_hours = [
            1,
            1,
        ]  #  input_data_dict['params']['passed_data']['def_total_hours']

        opt_res_naive_mpc = opt.perform_naive_mpc_optim(
            df_input_data_dayahead,
            P_PV_forecast,
            P_load_forecast,
            prediction_horizon,
            soc_init,
            soc_final,
            def_total_hours,
        )
        # Save CSV file for publish_data
        if save_data_to_file:
            today = datetime.now(timezone.utc).replace(hour=0, minute=0, second=0, microsecond=0)
            filename = "opt_res_naive_mpc_" + today.strftime("%Y_%m_%d") + ".csv"
        else:  # Just save the latest optimization results
            filename = "opt_res_naive_mpc_latest.csv"
        if not debug:
            opt_res_naive_mpc.to_csv(self._data_folder / filename, index_label="timestamp")
        return opt_res_naive_mpc

    def forecast_model_fit(
        self, only_if_file_does_not_exist: bool = False, days_to_retrieve: int | None = None
    ) -> float:
        """Perform a forecast model fit from training data retrieved from Home Assistant."""

        filename = LOAD_FORECAST_MODEL_TYPE + "_mlf.pkl"
        filename_path = self._data_folder / filename

        if only_if_file_does_not_exist and filename_path.is_file():
            self._logger.info("Skip model creation")
            return 0

        self._logger.info("Setting up needed data")

        # Treat runtimeparams
        params: str = ""
        params, RetrieveHass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            json.dumps(self.get_ml_runtime_params()),
            json.dumps(self._emhass_config),
            self._RetrieveHass_conf,
            self._optim_conf,
            self._plant_conf,
            "forecast-model-fit",
            self._logger,
        )  # type: ignore

        params_dict: dict = json.loads(params)
        # Retrieve data from hass
        if days_to_retrieve is None:
            days_to_retrieve = self._RetrieveHass_conf.get("days_to_retrieve", 10)

        days_list = utils.get_days_list(days_to_retrieve)
        var_list = [self._power_no_var_loads_id]
        self._RetrieveHass.get_data(days_list, var_list)
        df_input_data = self._RetrieveHass.df_final.copy()

        data = copy.deepcopy(df_input_data)
        model_type = params_dict["passed_data"]["model_type"]
        sklearn_model = params_dict["passed_data"]["sklearn_model"]
        num_lags = params_dict["passed_data"]["num_lags"]
        split_date_delta = params_dict["passed_data"]["split_date_delta"]
        perform_backtest = params_dict["passed_data"]["perform_backtest"]
        # The ML forecaster object
        mlf = MLForecaster(
            data,
            model_type,
            self._power_no_var_loads_id,
            sklearn_model,
            num_lags,
            str(self._data_folder),
            self._logger,
        )
        # Fit the ML model
        df_pred, df_pred_backtest = mlf.fit(
            split_date_delta=split_date_delta, perform_backtest=perform_backtest
        )
        predictions = df_pred["pred"].dropna()
        test_data = df_pred["test"].dropna()
        r2 = r2_score(test_data, predictions)
        self._logger.info(f"R2 score = {r2}")
        # Save model
        with open(self._data_folder / filename, "wb") as outp:
            pickle.dump(mlf, outp, pickle.HIGHEST_PROTOCOL)
        return r2

    def forecast_model_tune(self) -> Tuple[pd.DataFrame, MLForecaster]:
        """Tune a forecast model hyperparameters using bayesian optimization.

        :param debug: True to debug, useful for unit testing, defaults to False
        :type debug: bool, optional
        :return: The DataFrame containing the forecast data results using the optimized model
        :rtype: pd.DataFrame
        """
        # Load model
        self._logger.info("Tune the forecast model")

        filename = LOAD_FORECAST_MODEL_TYPE + "_mlf.pkl"
        filename_path = self._data_folder / filename
        if filename_path.is_file():
            with open(filename_path, "rb") as inp:
                mlf = pickle.load(inp)
            # Tune the model
            df_pred_optim = mlf.tune(debug=False)

            # Save model
            with open(filename_path, "wb") as outp:
                pickle.dump(mlf, outp, pickle.HIGHEST_PROTOCOL)
            return df_pred_optim, mlf

        else:
            self._logger.error(
                "The ML forecaster file was not found, please run a model fit method before this tune method"
            )
            raise Exception(
                "The ML forecaster file was not found, please run a model fit method before this tune method"
            )

    def forecast_model_predict(
        self,
        use_last_window: bool = True,
        debug: bool = False,
        mlf: MLForecaster | None = None,
    ) -> pd.Series | None:
        """Perform a forecast model predict using a previously trained skforecast model.

        :param input_data_dict: A dictionary with multiple data used by the action functions
        :type input_data_dict: dict
        :param logger: The passed logger object
        :type logger: logging.Logger
        :param use_last_window: True if the 'last_window' option should be used for the \
            custom machine learning forecast model. The 'last_window=True' means that the data \
            that will be used to generate the new forecast will be freshly retrieved from \
            Home Assistant. This data is needed because the forecast model is an auto-regressive \
            model with lags. If 'False' then the data using during the model train is used. Defaults to True
        :type use_last_window: Optional[bool], optional
        :param debug: True to debug, useful for unit testing, defaults to False
        :type debug: Optional[bool], optional
        :param mlf: The 'mlforecaster' object previously trained. This is mainly used for debug \
            and unit testing. In production the actual model will be read from a saved pickle file. Defaults to None
        :type mlf: Optional[mlforecaster], optional
        :return: The DataFrame containing the forecast prediction data
        :rtype: pd.DataFrame
        """
        # Treat runtimeparams
        params: str = ""
        params, RetrieveHass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            json.dumps(self.get_ml_runtime_params()),
            json.dumps(self._emhass_config),
            self._RetrieveHass_conf,
            self._optim_conf,
            self._plant_conf,
            "forecast-model-fit",
            self._logger,
        )  # type: ignore

        params_dict: dict = json.loads(params)
        # Retrieve data from hass
        days_to_retrieve = params_dict["passed_data"]["days_to_retrieve"]

        days_list = utils.get_days_list(days_to_retrieve)
        var_list = [self._power_no_var_loads_id]
        self._RetrieveHass.get_data(days_list, var_list)
        df_input_data = self._RetrieveHass.df_final.copy()

        # Load model
        model_type = "load_forecast"
        filename = model_type + "_mlf.pkl"
        filename_path = self._data_folder / filename
        if not debug:
            if filename_path.is_file():
                with open(filename_path, "rb") as inp:
                    mlf = pickle.load(inp)
            else:
                self._logger.error(
                    "The ML forecaster file was not found, please run a model fit method before this predict method"
                )
                return None
        # Make predictions
        if use_last_window:
            data_last_window = copy.deepcopy(df_input_data)
        else:
            data_last_window = None
        if mlf is not None:
            return mlf.predict(data_last_window)
        return None

    def get_forecast(self) -> ForecastSchema:
        """Get the previously calculated forecast."""
        temp_folder = self._data_folder / "temp"
        forecast_filename = "forecast.csv"

        if self._day_ahead_forecast is None:
            try:
                self._day_ahead_forecast = pd.read_csv(temp_folder / forecast_filename)
            except Exception:
                self._logger.exception(
                    f"{forecast_filename} is not available. Creating forecast..."
                )
                self.dayahead_forecast_optim()

        if self._day_ahead_forecast is not None:
            freq = self._RetrieveHass_conf["freq"]
            temp_folder.mkdir(parents=True, exist_ok=True)
            pv_df = self._pv.get_data_frame(freq, self._location.get_time_zone(), "pv", temp_folder)
            pv_df.to_csv(temp_folder / "pv_df.csv")
            no_var_load_df = self._no_var_loads.get_data_frame(
                freq, self._location.get_time_zone(), "non_var_loads", temp_folder
            )
            df = self._day_ahead_forecast.merge(
                pv_df, how="left", left_index=True, right_index=True
            )
            df = df.merge(no_var_load_df, how="left", left_index=True, right_index=True)

            df.rename(columns={"P_PV": "pv_forecast"}, inplace=True)
            df.to_csv(temp_folder / forecast_filename, index_label="time_stamp")

            while not pd.notnull(df["pv_forecast"][0]) and len(df.index) > 0:
                df.drop(df.index[0], inplace=True)

            pv_series = [x for x in df["pv"].to_list() if pd.notnull(x)]
            no_var_load_series = [x for x in df["non_var_loads"].to_list() if pd.notnull(x)]

            # TODO: Should be removed
            df.fillna(-10000, inplace=True)
            time_series = df.index.to_series()
            time: list[datetime] = time_series.tolist()
            pv_forecast = df["pv_forecast"].to_list()
            load = df["P_Load"].to_list()
            cost_profit = df["cost_profit"].to_list()

            series = [
                ForecastSerieSchema(name="pv_forecast", data=pv_forecast),
                ForecastSerieSchema(name="pv", data=pv_series),
                ForecastSerieSchema(name="consumption", data=load),
                ForecastSerieSchema(name="no_var_loads", data=no_var_load_series),
                ForecastSerieSchema(name="cost_profit", data=cost_profit),
            ]
            consumed_energy = sum(load)
            for i, d in enumerate(self._optimzed_devices):
                device = self._day_ahead_forecast[f"P_deferrable{i}"].to_list()
                series.append(ForecastSerieSchema(name=str(d.device_id), data=device))
                consumed_energy = consumed_energy + sum(device)

            period = freq.total_seconds() / 3600
            solar_energy = sum(pv_forecast) * period / 1000  # kWh
            consumed_energy = consumed_energy * period / 1000  # kWh
            cost = sum(cost_profit)

            return ForecastSchema(
                solar_energy=solar_energy,
                cost=cost,
                consumed_energy=consumed_energy,
                time=time,
                series=series,
            )
        else:
            raise Exception("Optimizer forecast is not initialized.")

    def get_optimized_power(self, device_id: uuid.UUID) -> float:
        """Get the optimized power budget for a give device."""
        if self._day_ahead_forecast is not None:
            for i, deferrable_load_info in enumerate(self._optimzed_devices):
                if deferrable_load_info.device_id == device_id:
                    columnName = f"P_deferrable{i}"
                    return self._get_forecast_value(columnName)
        return -1

    def _get_forecast_value(self, columnName: str) -> float:
        """Get a forecasted value."""
        if self._day_ahead_forecast is not None:
            now_precise = datetime.now(self._location.get_time_zone()).replace(
                second=0, microsecond=0
            )
            if self._method_ts_round == "nearest":
                method = "nearest"
            elif self._method_ts_round == "first":
                method = "ffill"
            elif self._method_ts_round == "last":
                method = "bfill"
            else:
                method = "nearest"

            idx_closest = self._day_ahead_forecast.index.get_indexer([now_precise], method=method)[  # type: ignore
                0
            ]
            if idx_closest == -1:
                idx_closest = self._day_ahead_forecast.index.get_indexer(  # type: ignore
                    [now_precise], method="nearest"
                )[0]

            value = self._day_ahead_forecast.iloc[idx_closest][columnName]
            return float(value)
        return -1

    def _has_deferrable_load(self, device_id: uuid.UUID) -> bool:
        for deferrable_load_info in self._optimzed_devices:
            if deferrable_load_info.device_id == device_id:
                return True
        return False

    def update_devices(self, home: Home) -> None:
        """Update the selected devices from the list of devices."""
        new_optimizhed_devices = []
        needs_update = False
        self._pv.add_data_point(home.solar_production_power)
        self._projected_load_devices.clear()
        for device in home.devices:
            load_info = device.get_load_info()
            if load_info is not None:
                if load_info.is_deferrable:
                    if not self._has_deferrable_load(device.id):
                        needs_update = True
                    new_optimizhed_devices.append(load_info)
                else:
                    self._projected_load_devices.append(load_info)
                    # TODO: Only if not already there
                    needs_update = True
            elif self._has_deferrable_load(device.id):
                needs_update = True
        self._optimzed_devices = new_optimizhed_devices
        if needs_update or self._day_ahead_forecast is None:
            self.dayahead_forecast_optim()



"""Handle meters which can loose their energy meter value and reset to 0. This is useful for example for meters which are resetting to 0 in case of power loss."""

import logging

from energy_assistant.constants import ROOT_LOGGER_NAME
from energy_assistant.devices import State

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)


class UtilityMeter:
    """Handle meters which can loose their energy meter value and reset to 0."""

    def __init__(self, name: str) -> None:
        """Create a utility meter instance."""
        self._last_meter_value: float = 0
        self._energy: float = 0
        self._name = name

    @property
    def name(self) -> str:
        """Name of the utility meter."""
        return self._name

    @property
    def energy(self) -> float:
        """Current energy."""
        return self._energy

    @property
    def last_meter_value(self) -> float:
        """The last measured meture value."""
        return self._last_meter_value

    def update_energy(self, energy: float) -> float:
        """Update the utility meter with a energy measurement."""
        delta: float = max(energy - self._last_meter_value, 0)
        self._last_meter_value = energy
        self._energy += delta
        return self._energy

    def update_energy_state(self, energy_state: State) -> State:
        """Update the utility meter with a energy measurement provide as State."""
        energy = energy_state.numeric_value if energy_state else 0.0
        return State(energy_state.id, str(self.update_energy(energy)), energy_state.attributes)

    def restore_last_meter_value(self, meter_value: float) -> None:
        """Restore the last meter value."""
        self._last_meter_value = meter_value



"""Tests for the data analysis classes."""

from datetime import datetime

import pandas as pd
import pytest

from energy_assistant.devices.analysis import DataBuffer, create_timeseries_from_const


@pytest.fixture
def power_data() -> DataBuffer:
    """Create power data test fixture."""
    result = DataBuffer()
    for x in range(0, 20):
        result.add_data_point(x, datetime(2023, 1, 10, 10, 10, x))
    return result


@pytest.fixture
def power_data_with_trailing_zeros() -> DataBuffer:
    """Create power data test fixture."""
    result = DataBuffer()
    for x in range(0, 20):
        result.add_data_point(x, datetime(2023, 1, 10, 10, 10, x))
    for y in range(0, 3):
        result.add_data_point(0.0, datetime(2023, 1, 10, 10, 10, 20 + y))
    return result


def test_min(power_data: DataBuffer):
    """Test the data buffer minimum calculation."""
    min = power_data.get_min_for(5, datetime(2023, 1, 10, 10, 10, 21))
    assert min == 16


def test_max(power_data: DataBuffer):
    """Test the data buffer maximum calculation."""
    max = power_data.get_max_for(5, datetime(2023, 1, 10, 10, 10, 21))
    assert max == 19


def test_average(power_data: DataBuffer):
    """Test the data buffer average calculation."""
    avg = power_data.get_average_for(5, datetime(2023, 1, 10, 10, 10, 21))
    assert avg == 17.5


def test_get_data_without_trailing_zeros(power_data_with_trailing_zeros: DataBuffer):
    """Test getting the data without the trailing zeros."""
    data = power_data_with_trailing_zeros.get_data_for(8, datetime(2023, 1, 10, 10, 10, 23), True)
    assert len(data) == 5
    assert data == [15, 16, 17, 18, 19]


def test_is_between_without_trailing_zeros(power_data_with_trailing_zeros: DataBuffer):
    """Test getting the data without the trailing zeros."""
    data = power_data_with_trailing_zeros.is_between(
        15, 19, 8, datetime(2023, 1, 10, 10, 10, 23), True
    )
    assert data


def test_create_timeseries_from_const():
    """Test creation of a constant time series."""
    data = create_timeseries_from_const(
        15, pd.Timedelta(60, "m"), pd.Timedelta(30, "s"), pd.Timestamp(2024, 1, 1, 12, 30, 15)
    )
    assert data.size == 121
    assert not data.hasnans
    assert data.min() == 15
    assert data.max() == 15



"""The home is the root of all devices."""

import logging
import uuid

from energy_assistant import Optimizer
from energy_assistant.constants import ROOT_LOGGER_NAME
from energy_assistant.devices.analysis import DataBuffer
from energy_assistant.devices.evcc import EvccDevice
from energy_assistant.devices.registry import DeviceTypeRegistry
from energy_assistant.devices.state_value import StateValue

from . import (
    HomeEnergySnapshot,
    SessionStorage,
    State,
    StatesRepository,
    assign_if_available,
)
from .config import DeviceConfigException, get_config_param
from .device import Device
from .heat_pump import HeatPumpDevice
from .homeassistant import HomeassistantDevice

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)


class Home:
    """The home."""

    def __init__(
        self,
        config: dict,
        session_storage: SessionStorage,
        device_type_registry: DeviceTypeRegistry,
    ) -> None:
        """Create a home instance."""
        self._name: str = get_config_param(config, "name")

        self._init_power_variables(config)

        self.grid_exported_power_data = DataBuffer()

        self._disable_device_control: bool = False
        disable_control = config.get("disable_device_control")
        if disable_control is not None and disable_control:
            self._disable_device_control = True

        self._energy_snapshop: HomeEnergySnapshot | None = None
        self._init_devices(config, session_storage, device_type_registry)

    def _init_devices(
        self,
        config: dict,
        session_storage: SessionStorage,
        device_type_registry: DeviceTypeRegistry,
    ) -> None:
        self.devices = list[Device]()
        config_devices = config.get("devices")
        if config_devices is not None:
            for config_device in config_devices:
                type = config_device.get("type")
                self.create_device(type, config_device, session_storage, device_type_registry)

    def create_device(
        self,
        type: str,
        config: dict,
        session_storage: SessionStorage,
        device_type_registry: DeviceTypeRegistry,
    ) -> None:
        """Create and add a new device."""
        if type == "homeassistant":
            self.devices.append(HomeassistantDevice(config, session_storage, device_type_registry))
        elif type == "heat-pump":
            self.devices.append(HeatPumpDevice(config, session_storage))
        elif type == "power-state-device":
            # This is deprecated and will be removed.
            self.devices.append(HomeassistantDevice(config, session_storage, device_type_registry))
        elif type == "evcc":
            self.devices.append(EvccDevice(config, session_storage))
        else:
            LOGGER.error(f"Unknown device type {type} in configuration")

    def _init_power_variables(self, config: dict) -> None:
        solar_power_config = config.get("solar_power")
        if solar_power_config is not None:
            self._solar_power_value = StateValue(solar_power_config)
        else:
            raise DeviceConfigException("Parameter solar_power is missing in the configuration")

        grid_supply_power_config = config.get("grid_supply_power")
        if grid_supply_power_config is not None:
            self._grid_imported_power_value = StateValue(grid_supply_power_config)
        else:
            raise DeviceConfigException("Parameter energy is missing in the configuration")

        grid_inverted: bool | None = config.get("grid_inverted")
        if grid_inverted is not None:
            if grid_inverted:
                self._grid_imported_power_value.invert_value()
            LOGGER.warning(
                "The home is configured with grid_inverted. This is deprecated and will no longer be supported."
            )

        solar_energy_config = config.get("solar_energy")
        if solar_energy_config is not None:
            self._solar_energy_value = StateValue(solar_energy_config)
        else:
            raise DeviceConfigException("Parameter solar_energy is missing in the configuration")

        imported_energy_config = config.get("imported_energy")
        if imported_energy_config is not None:
            self._imported_energy_value = StateValue(imported_energy_config)
        else:
            raise DeviceConfigException("Parameter imported_energy is missing in the configuration")

        exported_energy_config = config.get("exported_energy")
        if exported_energy_config is not None:
            self._exported_energy_value = StateValue(exported_energy_config)
        else:
            raise DeviceConfigException("Parameter exported_energy is missing in the configuration")

        self._solar_production_power: State | None = None
        self._grid_imported_power: State | None = None
        self._consumed_energy: float = 0.0
        self._consumed_solar_energy: float = 0.0

        self._grid_exported_energy: State | None = None
        self._grid_imported_energy: State | None = None
        self._produced_solar_energy: State | None = None

    def add_device(self, device: Device) -> None:
        """Add a device to the home."""
        self.devices.append(device)

    def remove_device(self, device_id: uuid.UUID) -> None:
        """Remove the device with a given id."""
        for device in self.devices:
            if device.id == device_id:
                self.devices.remove(device)
                return

    def get_device(self, id: uuid.UUID) -> Device | None:
        """Get device with the given id."""
        for device in self.devices:
            if device.id == id:
                return device
        return None

    @property
    def name(self) -> str:
        """Name of the Home."""
        return self._name

    @property
    def produced_solar_energy(self) -> float:
        """Solar energy in kWh."""
        return self._produced_solar_energy.numeric_value if self._produced_solar_energy else 0.0

    @property
    def grid_imported_energy(self) -> float:
        """Imported energy from the grid in kWh."""
        return self._grid_imported_energy.numeric_value if self._grid_imported_energy else 0.0

    @property
    def grid_exported_energy(self) -> float:
        """Exported energy from the grid in kWh."""
        return self._grid_exported_energy.numeric_value if self._grid_exported_energy else 0.0

    @property
    def consumed_energy(self) -> float:
        """Consumed energy in kWh."""
        return self._consumed_energy

    @property
    def consumed_solar_energy(self) -> float:
        """Consumed solar energy in kWh."""
        return self._consumed_solar_energy

    @property
    def home_consumption_power(self) -> float:
        """Consumpton power of the home."""
        result = self.solar_production_power - self.grid_imported_power
        if result > 0:
            return result
        else:
            return 0

    @property
    def solar_self_consumption_power(self) -> float:
        """Self consumed power of the home."""
        if self.grid_imported_power < 0:
            return self.solar_production_power
        else:
            return self.solar_production_power - self.grid_imported_power

    @property
    def self_sufficiency(self) -> float:
        """Self sufficiency of the home."""
        home_consumption = self.home_consumption_power
        if home_consumption > 0:
            return min(self.solar_self_consumption_power / home_consumption, 1.0)
        else:
            return 0

    @property
    def self_consumption(self) -> float:
        """Self consumption ratio of the home."""
        solar_power = self.solar_production_power
        if solar_power > 0:
            return min(self.home_consumption_power / solar_power, 1.0)
        else:
            return 0

    async def update_state(self, state_repository: StatesRepository) -> None:
        """Update the state of the home."""
        self._solar_production_power = assign_if_available(
            self._solar_production_power,
            self._solar_power_value.evaluate(state_repository),
        )
        self._grid_imported_power = assign_if_available(
            self._grid_imported_power,
            self._grid_imported_power_value.evaluate(state_repository),
        )

        self._produced_solar_energy = assign_if_available(
            self._produced_solar_energy,
            self._solar_energy_value.evaluate(state_repository),
        )
        self._grid_imported_energy = assign_if_available(
            self._grid_imported_energy,
            self._imported_energy_value.evaluate(state_repository),
        )
        self._grid_exported_energy = assign_if_available(
            self._grid_exported_energy,
            self._exported_energy_value.evaluate(state_repository),
        )

        self._consumed_energy = (
            self.grid_imported_energy - self.grid_exported_energy + self.produced_solar_energy
        )
        self._consumed_solar_energy = self.produced_solar_energy - self.grid_exported_energy

        if self._energy_snapshop is None:
            self.set_snapshot(
                self.consumed_solar_energy,
                self.consumed_energy,
                self.produced_solar_energy,
                self.grid_imported_energy,
                self.grid_exported_energy,
            )

        for device in self.devices:
            await device.update_state(state_repository, self.self_sufficiency)

    async def update_power_consumption(
        self, state_repository: StatesRepository, optimizer: Optimizer
    ) -> None:
        """Update the device based on the current pv availability."""
        self.grid_exported_power_data.add_data_point(self.grid_imported_power)

        if not self._disable_device_control:
            for device in self.devices:
                await device.update_power_consumption(
                    state_repository, optimizer, self.grid_exported_power_data
                )

    @property
    def icon(self) -> str:
        """The icon of the home."""
        return "mdi-home"

    @property
    def solar_production_power(self) -> float:
        """Solar production power of the home."""
        return self._solar_production_power.numeric_value if self._solar_production_power else 0.0

    @property
    def grid_imported_power(self) -> float:
        """Grid supply power of the home."""
        return self._grid_imported_power.numeric_value if self._grid_imported_power else 0.0

    def restore_state(
        self,
        consumed_solar_energy: float,
        consumed_energy: float,
        solar_produced_energy: float,
        grid_imported_energy: float,
        grid_exported_energy: float,
    ) -> None:
        """Restore the proviously stored state."""
        self._consumed_solar_energy = consumed_solar_energy
        self._consumed_energy = consumed_energy

        self._produced_solar_energy = State("", str(solar_produced_energy))
        self._grid_imported_energy = State("", str(grid_imported_energy))
        self._grid_exported_energy = State("", str(grid_exported_energy))

        self.set_snapshot(
            consumed_solar_energy,
            consumed_energy,
            solar_produced_energy,
            grid_imported_energy,
            grid_exported_energy,
        )

    def set_snapshot(
        self,
        consumed_solar_energy: float,
        consumed_energy: float,
        solar_produced_energy: float,
        grid_imported_energy: float,
        grid_exported_energy: float,
    ) -> None:
        """Set the energy snapshot for the home."""
        self._energy_snapshop = HomeEnergySnapshot(
            consumed_solar_energy,
            consumed_energy,
            solar_produced_energy,
            grid_imported_energy,
            grid_exported_energy,
        )

    def store_energy_snapshot(self) -> None:
        """Store the current values in the snapshot."""
        self.set_snapshot(
            self.consumed_solar_energy,
            self.consumed_energy,
            self.produced_solar_energy,
            self.grid_imported_energy,
            self.grid_exported_energy,
        )
        for device in self.devices:
            device.store_energy_snapshot()

    @property
    def energy_snapshop(self) -> HomeEnergySnapshot | None:
        """The last energy snapshot of the device."""
        return self._energy_snapshop



"""Data model and schema classes for a session logs."""

from __future__ import annotations

import uuid
from datetime import datetime
from typing import TYPE_CHECKING, AsyncIterator

from sqlalchemy import ForeignKey, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Mapped, mapped_column, relationship

if TYPE_CHECKING:
    from .device import Device

from .base import Base


class SessionLogEntry(Base):
    """Data model for a session log entry."""

    __tablename__ = "session_log"

    id: Mapped[int] = mapped_column(primary_key=True)
    text: Mapped[str]

    device_id: Mapped[uuid.UUID] = mapped_column(
        "device_id", ForeignKey("devices.id"), nullable=False
    )

    device: Mapped[Device] = relationship("Device", back_populates="session_log_entries")

    start: Mapped[datetime]
    end: Mapped[datetime]

    start_consumed_energy: Mapped[float]
    start_solar_consumed_energy: Mapped[float]

    end_consumed_energy: Mapped[float]
    end_solar_consumed_energy: Mapped[float]

    @classmethod
    async def read_all(cls, session: AsyncSession) -> AsyncIterator[SessionLogEntry]:
        """Read all session log entries."""
        stmt = select(cls)
        stream = await session.stream_scalars(stmt.order_by(cls.start.desc()))
        async for row in stream:
            yield row

    @classmethod
    async def read_by_device_id(
        cls, session: AsyncSession, device_id: uuid.UUID
    ) -> AsyncIterator[SessionLogEntry]:
        """Read all session log entries."""
        stmt = select(cls).where(cls.device_id == device_id)
        stream = await session.stream_scalars(stmt.order_by(cls.start.desc()))
        async for row in stream:
            yield row

    @classmethod
    async def read_by_id(cls, session: AsyncSession, id: int) -> SessionLogEntry | None:
        """Read a session log entry by id."""
        stmt = select(cls).where(cls.id == id)
        return await session.scalar(stmt.order_by(cls.id))

    @classmethod
    async def create(
        cls,
        session: AsyncSession,
        text: str,
        device_id: uuid.UUID,
        start: datetime,
        start_solar_consumed_energy: float,
        start_consumed_energy: float,
        end: datetime,
        end_solar_consumed_energy: float,
        end_consumed_energy: float,
    ) -> SessionLogEntry:
        """Create a session log entry."""
        device = SessionLogEntry(
            text=text,
            device_id=device_id,
            start=start,
            start_solar_consumed_energy=start_solar_consumed_energy,
            start_consumed_energy=start_consumed_energy,
            end=end,
            end_solar_consumed_energy=end_solar_consumed_energy,
            end_consumed_energy=end_consumed_energy,
        )
        session.add(device)
        await session.flush()
        return device

    async def update(
        self,
        session: AsyncSession,
        text: str,
        device_id: uuid.UUID,
        start: datetime,
        start_solar_consumed_energy: float,
        start_consumed_energy: float,
        end: datetime,
        end_solar_consumed_energy: float,
        end_consumed_energy: float,
    ) -> None:
        """Update a session log entry."""
        self.text = text
        self.device_id = device_id
        self.start = start
        self.start_solar_consumed_energy = start_solar_consumed_energy
        self.start_consumed_energy = start_consumed_energy
        self.end = end
        self.end_solar_consumed_energy = end_solar_consumed_energy
        self.end_consumed_energy = end_consumed_energy
        await session.flush()

    @classmethod
    async def delete(cls, session: AsyncSession, session_log_entry: SessionLogEntry) -> None:
        """Delete a session log entry."""
        await session.delete(session_log_entry)
        await session.flush()



"""Views for home measurement API."""

from fastapi import APIRouter, Depends, HTTPException, Request

from .schema import ConfigModel, ReadConfigResponse
from .use_cases import ReadConfiguration

router = APIRouter(prefix="/config")


@router.get("", response_model=ReadConfigResponse)
async def read_configuration(
    request: Request,
    use_case: ReadConfiguration = Depends(ReadConfiguration),
) -> ConfigModel:
    """Rest end point for read all devices."""
    energy_assistant = (
        request.app.energy_assistant if hasattr(request.app, "energy_assistant") else None
    )
    if energy_assistant is None:
        raise HTTPException(status_code=500)
    return await use_case.execute(energy_assistant.config)



"""Use cases for home measurements."""

from datetime import date
from typing import AsyncIterator

from fastapi import HTTPException

from energy_assistant.db import AsyncSession
from energy_assistant.models.home import HomeMeasurement
from energy_assistant.models.schema import HomeMeasurementSchema


class ReadAllHomeMeasurement:
    """Read all home measurements use case."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a read all home measurements use case."""
        self.async_session = session

    async def execute(self) -> AsyncIterator[HomeMeasurementSchema]:
        """Execute the read all home measurements use case."""
        async with self.async_session() as session:
            async for home_measurement in HomeMeasurement.read_all(
                session, include_device_measurements=True
            ):
                yield HomeMeasurementSchema.model_validate(home_measurement)


class ReadHomeMeasurement:
    """Read a home measurement use case."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a read home measurement use case."""
        self.async_session = session

    async def execute(self, home_measurement_id: int) -> HomeMeasurementSchema:
        """Execute the read home measurement use case."""
        async with self.async_session() as session:
            home_measurement = await HomeMeasurement.read_by_id(
                session, home_measurement_id, include_device_measurements=True
            )
            if not home_measurement:
                raise HTTPException(status_code=404)
            return HomeMeasurementSchema.model_validate(home_measurement)


class ReadHomeMeasurementByDate:
    """Read a home measurement by date use case."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a read home measurement use case."""
        self.async_session = session

    async def execute(self, measurement_date: date) -> HomeMeasurementSchema:
        """Execute the read home measurement use case."""
        async with self.async_session() as session:
            home_measurement = await HomeMeasurement.read_by_date(
                session, measurement_date, include_device_measurements=True
            )
            if not home_measurement:
                raise HTTPException(status_code=404)
            return HomeMeasurementSchema.model_validate(home_measurement)


class ReadHomeMeasurementLastBeforeDate:
    """Read the last home measurement before a date use case."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a read home measurement use case."""
        self.async_session = session

    async def execute(self, measurement_date: date) -> HomeMeasurementSchema:
        """Execute the read home measurement use case."""
        async with self.async_session() as session:
            home_measurement = await HomeMeasurement.read_before_date(
                session, measurement_date, include_device_measurements=True
            )
            if not home_measurement:
                raise HTTPException(status_code=404)
            return HomeMeasurementSchema.model_validate(home_measurement)


class DeleteHomeMeasurement:
    """Delete a home measurement use case."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a delete a home measurement use case."""
        self.async_session = session

    async def execute(self, home_heasurement_id: int) -> None:
        """Execute the delete a home measurement use case."""
        async with self.async_session.begin() as session:
            home_measurement = await HomeMeasurement.read_by_id(session, home_heasurement_id)
            if not home_measurement:
                return
            await HomeMeasurement.delete(session, home_measurement)



"""Base class for Energy Assistant data models."""

import uuid
from typing import Mapping

from sqlalchemy import MetaData
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.types import CHAR, TypeDecorator

convention: Mapping[str, str] = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}


class GUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    """

    impl = CHAR
    cache_ok = True

    def load_dialect_impl(self, dialect):  # type:ignore
        """Load the dialect implementation."""
        if dialect.name == "postgresql":
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):  # type:ignore
        """Process the bind parameters."""
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return str(value)
        elif not isinstance(value, uuid.UUID):
            return "%.32x" % uuid.UUID(value).int
        else:
            # hexstring
            return "%.32x" % value.int

    def process_result_value(self, value, dialect):  # type:ignore
        """Process the result value."""
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value


class Base(AsyncAttrs, DeclarativeBase):
    """Base class for energy assistant data models."""

    __abstract__ = True
    metadata = MetaData(naming_convention=convention)  # type: ignore

    type_annotation_map = {
        uuid.UUID: GUID,
    }

    def __repr__(self) -> str:
        """Representation of a data model object."""
        columns = ", ".join(
            [f"{k}={repr(v)}" for k, v in self.__dict__.items() if not k.startswith("_")]
        )
        return f"<{self.__class__.__name__}({columns})>"



"""Schemas for session logs api."""

import uuid
from datetime import datetime

from pydantic import BaseModel, ConfigDict


class SessionLogEntrySchema(BaseModel):
    """API Response for reading a session log entry."""

    start: datetime
    end: datetime
    text: str
    device_id: uuid.UUID
    solar_consumed_energy: float
    consumed_energy: float

    model_config = ConfigDict(from_attributes=True)


class ReadAllSessionLogEntriesResponse(BaseModel):
    """API Response for reading session log entries."""

    entries: list[SessionLogEntrySchema]



"""The Device classes."""

import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime, tzinfo
from enum import StrEnum, auto
from typing import Any, Optional

import pytz  # type: ignore


@dataclass
class LoadInfo:
    """Information about a deferrable load."""

    device_id: uuid.UUID
    nominal_power: float
    duration: float
    is_continous: bool
    is_constant: bool = False
    start_timestep: int = 0
    end_timestep: int = 0
    is_deferrable: bool = True


@dataclass
class Session:
    """Session data class."""

    id: int
    start: datetime
    start_solar_consumed_energy: float
    start_consumed_energy: float


class SessionStorage(ABC):
    """Session storage base class."""

    @abstractmethod
    async def start_session(
        self,
        device_id: uuid.UUID,
        text: str,
        solar_consumed_energy: float,
        consumed_energy: float,
    ) -> Session:
        """Start a new session."""
        pass

    @abstractmethod
    async def update_session(
        self, id: int, solar_consumed_energy: float, consumed_energy: float
    ) -> None:
        """Update the session with the given id."""
        pass

    @abstractmethod
    async def update_session_energy(
        self, id: int, solar_consumed_energy: float, consumed_energy: float
    ) -> None:
        """Update the session with the given id."""
        pass


class Integrator:
    """Integrate a measurement like power to get the energy."""

    def __init__(self) -> None:
        """Initialize the integrator."""
        self.last_measurement: Optional[float] = None
        self.last_timestamp: Optional[float] = None
        self._value: float = 0.0

    @property
    def value(self) -> float:
        """The current value of the integrator."""
        return self._value

    def add_measurement(self, measurement: float, timestamp: float) -> None:
        """Update the value of the integrator with and new measuremenent value."""
        if self.last_measurement is None:
            self.last_measurement = measurement
            self.last_timestamp = timestamp
        else:
            delta_t = timestamp - self.last_timestamp if self.last_timestamp else 0.0
            self.last_timestamp = timestamp
            # print("Delta t: "+ str(delta_t))
            if delta_t > 0.1:
                if measurement > self.last_measurement:
                    self._value = self._value + (
                        delta_t
                        * (self.last_measurement + (measurement - self.last_measurement) / 2)
                    )
                else:
                    self._value = self._value + (
                        delta_t * (measurement + (self.last_measurement - measurement) / 2)
                    )

    def restore_state(self, state: float) -> None:
        """Restore the integrator value with a previously saved state."""
        self._value = state


class EnergyIntegrator:
    """Integrates energy based on a real measurement and a self sufficiency value in order to sum up the consumed solar energy."""

    def __init__(self) -> None:
        """Create an energy integrator."""
        self._last_consumed_energy: float = 0.0
        self._consumed_solar_energy: float = 0.0

    @property
    def consumed_solar_energy(self) -> float:
        """The amount of solar energy which has been consumed."""
        return self._consumed_solar_energy

    def add_measurement(self, consumed_energy: float, self_sufficiency: float) -> None:
        """Update the value of the integrator with and new measuremenent value."""
        self._consumed_solar_energy = (
            self._consumed_solar_energy
            + (consumed_energy - self._last_consumed_energy) * self_sufficiency
        )
        self._last_consumed_energy = consumed_energy

    def restore_state(self, consumed_solar_energy: float, last_consumed_energy: float) -> None:
        """Restores the integrator value with a previously saved state."""
        self._consumed_solar_energy = consumed_solar_energy
        self._last_consumed_energy = last_consumed_energy


class EnergySnapshot:
    """Stores an snapshot of the current energy consumption values."""

    def __init__(self, consumed_solar_energy: float, consumed_energy: float) -> None:
        """Create an energy snapshot."""
        self._consumed_solar_energy = consumed_solar_energy
        self._consumed_energy = consumed_energy

    @property
    def consumed_solar_energy(self) -> float:
        """The amount of consumed solar energy."""
        return self._consumed_solar_energy

    @property
    def consumed_energy(self) -> float:
        """The total amount of consumed energy."""
        return self._consumed_energy


class HomeEnergySnapshot(EnergySnapshot):
    """Stores an snapshot of the current energy consumption values."""

    def __init__(
        self,
        consumed_solar_energy: float,
        consumed_energy: float,
        solar_produced_energy: float,
        grid_imported_energy: float,
        grid_exported_energy: float,
    ) -> None:
        """Create an energy snapshot."""
        self._consumed_solar_energy = consumed_solar_energy
        self._consumed_energy = consumed_energy
        self._produced_solar_energy = solar_produced_energy
        self._grid_imported_energy = grid_imported_energy
        self._grid_exported_energy = grid_exported_energy

    @property
    def consumed_solar_energy(self) -> float:
        """The amount of consumed solar energy."""
        return self._consumed_solar_energy

    @property
    def consumed_energy(self) -> float:
        """The total amount of consumed energy."""
        return self._consumed_energy

    @property
    def produced_solar_energy(self) -> float:
        """The total amount of produced solar energy."""
        return self._produced_solar_energy

    @property
    def grid_imported_energy(self) -> float:
        """The total amount of energy imported from the grid."""
        return self._grid_imported_energy

    @property
    def grid_exported_energy(self) -> float:
        """The total amount of energy exported to the grid."""
        return self._grid_exported_energy


class OnOffState(StrEnum):
    """Representation of a on/off state."""

    ON = "on"
    OFF = "off"
    UNKNOWN = "unknown"


class State:
    """Base class for States."""

    def __init__(self, id: str, value: str, attributes: dict = {}) -> None:
        """Create a state instance."""
        self._id = id
        self._value = value
        self._attributes = attributes
        self._available = True

    @property
    def id(self) -> str:
        """Id of the state."""
        return self._id

    @property
    def available(self) -> bool:
        """Availability of the state."""
        return self._available

    @property
    def value(self) -> str:
        """State of the state as string."""
        return self._value

    @property
    def numeric_value(self) -> float:
        """Numeric state of the state."""
        try:
            return float(self._value)
        except ValueError:
            return 0.0

    @property
    def attributes(self) -> dict:
        """Attributes of the state."""
        return self._attributes


def assign_if_available(old_state: State | None, new_state: State | None) -> State | None:
    """Return new state in case the state is available, otherwise old state."""
    if new_state and new_state.available:
        return new_state
    else:
        return old_state


@dataclass(frozen=True, eq=True)
class StateId:
    """The id of a state."""

    id: str
    channel: str


class StatesRepository(ABC):
    """Abstract base class for a state repositiroy."""

    @abstractmethod
    def get_state(self, id: StateId | str) -> State | None:
        """Get a state from the repository."""
        pass

    @abstractmethod
    def get_numeric_states(self) -> dict[str, float]:
        """Get a states from the repository."""
        pass

    @abstractmethod
    def get_template_states(self) -> dict:
        """Get a states from the repository."""
        pass

    @abstractmethod
    def set_state(self, id: StateId, value: str, attributes: dict = {}) -> None:
        """Set a state in the repository."""
        pass

    @property
    @abstractmethod
    def channel(self) -> str:
        """Get the channel of the State Repository."""
        pass

    @abstractmethod
    def read_states(self) -> None:
        """Read the states from the channel."""
        pass

    @abstractmethod
    def write_states(self) -> None:
        """Write the states to the channel."""
        pass


class StatesSingleRepository(StatesRepository):
    """Base class for a state repositiroy."""

    def __init__(self, channel: str) -> None:
        """Create a StatesRepository instance."""
        self._channel = channel
        self._read_states: dict[str, State] = dict[str, State]()
        self._write_states: dict[str, State] = dict[str, State]()
        self._template_states: dict[str, dict | Any] | None = None

    def get_state(self, id: StateId | str) -> State | None:
        """Get a state from the repository."""
        if isinstance(id, str):
            return self._read_states.get(id)
        else:
            return self._read_states.get(id.id)

    def get_numeric_states(self) -> dict[str, float]:
        """Get a states from the repository."""
        result = {k: v.numeric_value for k, v in self._read_states.items()}
        return result

    def get_template_states(self) -> dict:
        """Get template states from the repository."""
        if self._template_states is None:
            self._template_states = {}
            states = self.get_numeric_states().items()
            for k, v in states:
                parts = k.split(".")
                if len(parts) > 1:
                    type = parts[0]
                    attribute = parts[1]
                    if type in self._template_states:
                        self._template_states[type][attribute] = v
                    else:
                        self._template_states[type] = {attribute: v}
                else:
                    self._template_states[k] = v
        return self._template_states

    def set_state(self, id: StateId, value: str, attributes: dict = {}) -> None:
        """Set a state in the repository."""
        self._write_states[id.id] = State(id.id, value, attributes)

    @property
    def channel(self) -> str:
        """Get the channel of the State Repository."""
        return self._channel


class StatesMultipleRepositories(StatesRepository):
    """Base class for a state repositiroy."""

    def __init__(self, repositories: list[StatesRepository]) -> None:
        """Create a StatesRepository instance."""
        self._repositories = repositories

    def get_state(self, id: StateId | str) -> State | None:
        """Get a state from the repository."""
        if isinstance(id, StateId):
            for repository in self._repositories:
                if repository.channel == id.channel:
                    result = repository.get_state(id)
                    if result is not None:
                        return result
        else:
            for repository in self._repositories:
                result = repository.get_state(id)
                if result is not None:
                    return result
        return None

    def get_numeric_states(self) -> dict[str, float]:
        """Get a states from the repository."""
        result: dict[str, float] = {}
        for reposititory in self._repositories:
            result = {**result, **reposititory.get_numeric_states()}
        return result

    def get_template_states(self) -> dict:
        """Get template states from the repository."""
        result: dict = {}
        for reposititory in self._repositories:
            result = {**result, **reposititory.get_template_states()}
        return result

    def set_state(self, id: StateId, value: str, attributes: dict = {}) -> None:
        """Set a state in the repository."""
        for repository in self._repositories:
            if id.channel == repository.channel:
                repository.set_state(id, value, attributes)

    @property
    def channel(self) -> str:
        """Get the channel of the State Repository."""
        return "multiple"

    def read_states(self) -> None:
        """Read the states from the channel."""
        for repository in self._repositories:
            repository.read_states()

    def write_states(self) -> None:
        """Write the states to the channel."""
        for repository in self._repositories:
            repository.write_states()


class PowerModes(StrEnum):
    """Power modes for controlling the device."""

    DEVICE_CONTROLLED = auto()
    OFF = auto()
    PV = auto()
    MIN_PV = auto()
    FAST = auto()
    OPTIMIZED = auto()


@dataclass
class Location:
    """Location of the home."""

    time_zone: str
    latitude: str
    longitude: str
    elevation: str

    def get_time_zone(self) -> tzinfo:
        """Get the timezone."""
        return pytz.timezone(self.time_zone)
